# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module is to contain an improved bounding box.
"""

import abc
import copy
import warnings
from collections.abc import Callable
from typing import Any, NamedTuple, Self

import numpy as np

from astropy.units import Quantity, UnitBase
from astropy.utils.compat import COPY_IF_NEEDED

__all__ = ["CompoundBoundingBox", "ModelBoundingBox"]


class _BaseInterval(NamedTuple):
    lower: float
    upper: float


class _Interval(_BaseInterval):
    """
    A single input's bounding box interval.

    Parameters
    ----------
    lower : float
        The lower bound of the interval

    upper : float
        The upper bound of the interval

    Methods
    -------
    validate :
        Constructs a valid interval

    outside :
        Determine which parts of an input array are outside the interval.

    domain :
        Constructs a discretization of the points inside the interval.
    """

    def __repr__(self):
        return f"Interval(lower={self.lower}, upper={self.upper})"

    def copy(self):
        return copy.deepcopy(self)

    @staticmethod
    def _validate_shape(interval):
        """Validate the shape of an interval representation."""
        MESSAGE = """An interval must be some sort of sequence of length 2"""

        try:
            shape = np.shape(interval)
        except TypeError:
            try:
                # np.shape does not work with lists of Quantities
                if len(interval) == 1:
                    interval = interval[0]
                shape = np.shape([b.to_value() for b in interval])
            except (ValueError, TypeError, AttributeError):
                raise ValueError(MESSAGE)

        valid_shape = shape in ((2,), (1, 2), (2, 0))
        if not valid_shape:
            valid_shape = (
                len(shape) > 0
                and shape[0] == 2
                and all(isinstance(b, np.ndarray) for b in interval)
            )

        if not np.iterable(interval) or not valid_shape:
            raise ValueError(MESSAGE)

    @classmethod
    def _validate_bounds(cls, lower, upper):
        """Validate the bounds are reasonable and construct an interval from them."""
        if (np.asanyarray(lower) > np.asanyarray(upper)).all():
            warnings.warn(
                f"Invalid interval: upper bound {upper} "
                f"is strictly less than lower bound {lower}.",
                RuntimeWarning,
            )

        return cls(lower, upper)

    @classmethod
    def validate(cls, interval):
        """
        Construct and validate an interval.

        Parameters
        ----------
        interval : iterable
            A representation of the interval.

        Returns
        -------
        A validated interval.
        """
        cls._validate_shape(interval)

        if len(interval) == 1:
            interval = tuple(interval[0])
        else:
            interval = tuple(interval)

        return cls._validate_bounds(interval[0], interval[1])

    def outside(self, _input: np.ndarray):
        """
        Parameters
        ----------
        _input : np.ndarray
            The evaluation input in the form of an array.

        Returns
        -------
        Boolean array indicating which parts of _input are outside the interval:
            True  -> position outside interval
            False -> position inside  interval
        """
        return np.logical_or(_input < self.lower, _input > self.upper)

    def domain(self, resolution):
        return np.arange(self.lower, self.upper + resolution, resolution)


# The interval where all ignored inputs can be found.
_ignored_interval = _Interval.validate((-np.inf, np.inf))


def get_index(model, key) -> int:
    """
    Get the input index corresponding to the given key.
        Can pass in either:
            the string name of the input or
            the input index itself.
    """
    if isinstance(key, str):
        if key in model.inputs:
            index = model.inputs.index(key)
        else:
            raise ValueError(f"'{key}' is not one of the inputs: {model.inputs}.")
    elif np.issubdtype(type(key), np.integer):
        if 0 <= key < len(model.inputs):
            index = key
        else:
            raise IndexError(
                f"Integer key: {key} must be non-negative and < {len(model.inputs)}."
            )
    else:
        raise ValueError(f"Key value: {key} must be string or integer.")

    return index


def get_name(model, index: int):
    """Get the input name corresponding to the input index."""
    return model.inputs[index]


class _BoundingDomain(abc.ABC):
    """
    Base class for ModelBoundingBox and CompoundBoundingBox.
        This is where all the `~astropy.modeling.core.Model` evaluation
        code for evaluating with a bounding box is because it is common
        to both types of bounding box.

    Parameters
    ----------
    model : `~astropy.modeling.Model`
        The Model this bounding domain is for.

    prepare_inputs :
        Generates the necessary input information so that model can
        be evaluated only for input points entirely inside bounding_box.
        This needs to be implemented by a subclass. Note that most of
        the implementation is in ModelBoundingBox.

    prepare_outputs :
        Fills the output values in for any input points outside the
        bounding_box.

    evaluate :
        Performs a complete model evaluation while enforcing the bounds
        on the inputs and returns a complete output.
    """

    def __init__(self, model, ignored: list[int] | None = None, order: str = "C"):
        self._model = model
        self._ignored = self._validate_ignored(ignored)
        self._order = self._get_order(order)

    @property
    def model(self):
        return self._model

    @property
    def order(self) -> str:
        return self._order

    @property
    def ignored(self) -> list[int]:
        return self._ignored

    def _get_order(self, order: str | None = None) -> str:
        """
        Get if bounding_box is C/python ordered or Fortran/mathematically
        ordered.
        """
        if order is None:
            order = self._order

        if order not in ("C", "F"):
            raise ValueError(
                "order must be either 'C' (C/python order) or "
                f"'F' (Fortran/mathematical order), got: {order}."
            )

        return order

    def _get_index(self, key) -> int:
        """
        Get the input index corresponding to the given key.
            Can pass in either:
                the string name of the input or
                the input index itself.
        """
        return get_index(self._model, key)

    def _get_name(self, index: int):
        """Get the input name corresponding to the input index."""
        return get_name(self._model, index)

    @property
    def ignored_inputs(self) -> list[str]:
        return [self._get_name(index) for index in self._ignored]

    def _validate_ignored(self, ignored: list) -> list[int]:
        if ignored is None:
            return []
        else:
            return [self._get_index(key) for key in ignored]

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(
            "This bounding box is fixed by the model and does not have "
            "adjustable parameters."
        )

    @abc.abstractmethod
    def fix_inputs(self, model, fixed_inputs: dict):
        """
        Fix the bounding_box for a `fix_inputs` compound model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The new model for which this will be a bounding_box
        fixed_inputs : dict
            Dictionary of inputs which have been fixed by this bounding box.
        """
        raise NotImplementedError("This should be implemented by a child class.")

    @abc.abstractmethod
    def prepare_inputs(self, input_shape, inputs) -> tuple[Any, Any, Any]:
        """
        Get prepare the inputs with respect to the bounding box.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        inputs : list
            List of all the model inputs

        Returns
        -------
        valid_inputs : list
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : array_like
            array of all indices inside the bounding box
        all_out: bool
            if all of the inputs are outside the bounding_box
        """
        raise NotImplementedError("This has not been implemented for BoundingDomain.")

    @staticmethod
    def _base_output(input_shape, fill_value):
        """
        Create a baseline output, assuming that the entire input is outside
        the bounding box.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box

        Returns
        -------
        An array of the correct shape containing all fill_value
        """
        return np.zeros(input_shape) + fill_value

    def _all_out_output(self, input_shape, fill_value):
        """
        Create output if all inputs are outside the domain.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box

        Returns
        -------
        A full set of outputs for case that all inputs are outside domain.
        """
        return [
            self._base_output(input_shape, fill_value)
            for _ in range(self._model.n_outputs)
        ], None

    def _modify_output(self, valid_output, valid_index, input_shape, fill_value):
        """
        For a single output fill in all the parts corresponding to inputs
        outside the bounding box.

        Parameters
        ----------
        valid_output : numpy array
            The output from the model corresponding to inputs inside the
            bounding box
        valid_index : numpy array
            array of all indices of inputs inside the bounding box
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box

        Returns
        -------
        An output array with all the indices corresponding to inputs
        outside the bounding box filled in by fill_value
        """
        output = self._base_output(input_shape, fill_value)
        if not output.shape:
            output = np.array(valid_output)
        else:
            output[valid_index] = valid_output

        if np.isscalar(valid_output):
            output = output.item(0)

        return output

    def _prepare_outputs(self, valid_outputs, valid_index, input_shape, fill_value):
        """
        Fill in all the outputs of the model corresponding to inputs
        outside the bounding_box.

        Parameters
        ----------
        valid_outputs : list of numpy array
            The list of outputs from the model corresponding to inputs
            inside the bounding box
        valid_index : numpy array
            array of all indices of inputs inside the bounding box
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box

        Returns
        -------
        List of filled in output arrays.
        """
        outputs = []
        for valid_output in valid_outputs:
            outputs.append(
                self._modify_output(valid_output, valid_index, input_shape, fill_value)
            )

        return outputs

    def prepare_outputs(self, valid_outputs, valid_index, input_shape, fill_value):
        """
        Fill in all the outputs of the model corresponding to inputs
        outside the bounding_box, adjusting any single output model so that
        its output becomes a list of containing that output.

        Parameters
        ----------
        valid_outputs : list
            The list of outputs from the model corresponding to inputs
            inside the bounding box
        valid_index : array_like
            array of all indices of inputs inside the bounding box
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box
        """
        if self._model.n_outputs == 1:
            valid_outputs = [valid_outputs]

        return self._prepare_outputs(
            valid_outputs, valid_index, input_shape, fill_value
        )

    @staticmethod
    def _get_valid_outputs_unit(valid_outputs, with_units: bool) -> UnitBase | None:
        """
        Get the unit for outputs if one is required.

        Parameters
        ----------
        valid_outputs : list of numpy array
            The list of outputs from the model corresponding to inputs
            inside the bounding box
        with_units : bool
            whether or not a unit is required
        """
        if with_units:
            return getattr(valid_outputs, "unit", None)

    def _evaluate_model(
        self,
        evaluate: Callable,
        valid_inputs,
        valid_index,
        input_shape,
        fill_value,
        with_units: bool,
    ):
        """
        Evaluate the model using the given evaluate routine.

        Parameters
        ----------
        evaluate : Callable
            callable which takes in the valid inputs to evaluate model
        valid_inputs : list of numpy arrays
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : numpy array
            array of all indices inside the bounding box
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box
        with_units : bool
            whether or not a unit is required

        Returns
        -------
        outputs :
            list containing filled in output values
        valid_outputs_unit :
            the unit that will be attached to the outputs
        """
        valid_outputs = evaluate(valid_inputs)
        valid_outputs_unit = self._get_valid_outputs_unit(valid_outputs, with_units)

        return (
            self.prepare_outputs(valid_outputs, valid_index, input_shape, fill_value),
            valid_outputs_unit,
        )

    def _evaluate(
        self, evaluate: Callable, inputs, input_shape, fill_value, with_units: bool
    ):
        """Evaluate model with steps: prepare_inputs -> evaluate -> prepare_outputs.

        Parameters
        ----------
        evaluate : Callable
            callable which takes in the valid inputs to evaluate model
        valid_inputs : list of numpy arrays
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : numpy array
            array of all indices inside the bounding box
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box
        with_units : bool
            whether or not a unit is required

        Returns
        -------
        outputs :
            list containing filled in output values
        valid_outputs_unit :
            the unit that will be attached to the outputs
        """
        valid_inputs, valid_index, all_out = self.prepare_inputs(input_shape, inputs)

        if all_out:
            return self._all_out_output(input_shape, fill_value)
        else:
            return self._evaluate_model(
                evaluate, valid_inputs, valid_index, input_shape, fill_value, with_units
            )

    @staticmethod
    def _set_outputs_unit(outputs, valid_outputs_unit):
        """
        Set the units on the outputs
            prepare_inputs -> evaluate -> prepare_outputs -> set output units.

        Parameters
        ----------
        outputs :
            list containing filled in output values
        valid_outputs_unit :
            the unit that will be attached to the outputs

        Returns
        -------
        List containing filled in output values and units
        """
        if valid_outputs_unit is not None:
            return Quantity(
                outputs, valid_outputs_unit, copy=COPY_IF_NEEDED, subok=True
            )

        return outputs

    def evaluate(self, evaluate: Callable, inputs, fill_value):
        """
        Perform full model evaluation steps:
            prepare_inputs -> evaluate -> prepare_outputs -> set output units.

        Parameters
        ----------
        evaluate : callable
            callable which takes in the valid inputs to evaluate model
        valid_inputs : list
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : array_like
            array of all indices inside the bounding box
        fill_value : float
            The value which will be assigned to inputs which are outside
            the bounding box
        """
        input_shape = self._model.input_shape(inputs)

        # NOTE: CompoundModel does not currently support units during
        #   evaluation for bounding_box so this feature is turned off
        #   for CompoundModel(s).
        outputs, valid_outputs_unit = self._evaluate(
            evaluate, inputs, input_shape, fill_value, self._model.bbox_with_units
        )
        return tuple(self._set_outputs_unit(outputs, valid_outputs_unit))


class ModelBoundingBox(_BoundingDomain):
    """
    A model's bounding box.

    Parameters
    ----------
    intervals : dict
        A dictionary containing all the intervals for each model input
            keys   -> input index
            values -> interval for that index

    model : `~astropy.modeling.Model`
        The Model this bounding_box is for.

    ignored : list
        A list containing all the inputs (index) which will not be
        checked for whether or not their elements are in/out of an interval.

    order : optional, str
        The ordering that is assumed for the tuple representation of this
        bounding_box. Options: 'C': C/Python order, e.g. z, y, x.
        (default), 'F': Fortran/mathematical notation order, e.g. x, y, z.
    """

    def __init__(
        self,
        intervals: dict[int, _Interval],
        model,
        ignored: list[int] | None = None,
        order: str = "C",
    ):
        super().__init__(model, ignored, order)

        self._intervals = {}
        if intervals != () and intervals != {}:
            self._validate(intervals, order=order)

    def copy(self, ignored=None):
        intervals = {
            index: interval.copy() for index, interval in self._intervals.items()
        }

        if ignored is None:
            ignored = self._ignored.copy()

        return ModelBoundingBox(
            intervals, self._model, ignored=ignored, order=self._order
        )

    @property
    def intervals(self) -> dict[int, _Interval]:
        """Return bounding_box labeled using input positions."""
        return self._intervals

    @property
    def named_intervals(self) -> dict[str, _Interval]:
        """Return bounding_box labeled using input names."""
        return {self._get_name(index): bbox for index, bbox in self._intervals.items()}

    def __repr__(self):
        parts = ["ModelBoundingBox(", "    intervals={"]

        for name, interval in self.named_intervals.items():
            parts.append(f"        {name}: {interval}")

        parts.append("    }")
        if len(self._ignored) > 0:
            parts.append(f"    ignored={self.ignored_inputs}")

        parts.append(
            f"    model={self._model.__class__.__name__}(inputs={self._model.inputs})"
        )
        parts.append(f"    order='{self._order}'")
        parts.append(")")

        return "\n".join(parts)

    def __len__(self):
        return len(self._intervals)

    def __contains__(self, key):
        try:
            return self._get_index(key) in self._intervals or self._ignored
        except (IndexError, ValueError):
            return False

    def has_interval(self, key):
        return self._get_index(key) in self._intervals

    def __getitem__(self, key):
        """Get bounding_box entries by either input name or input index."""
        index = self._get_index(key)
        if index in self._ignored:
            return _ignored_interval
        else:
            return self._intervals[self._get_index(key)]

    def bounding_box(
        self, order: str | None = None
    ) -> tuple[float, float] | tuple[tuple[float, float], ...]:
        """
        Return the old tuple of tuples representation of the bounding_box
            order='C' corresponds to the old bounding_box ordering
            order='F' corresponds to the gwcs bounding_box ordering.
        """
        if len(self._intervals) == 1:
            return tuple(next(iter(self._intervals.values())))
        else:
            order = self._get_order(order)
            inputs = self._model.inputs
            if order == "C":
                inputs = inputs[::-1]

            bbox = tuple(tuple(self[input_name]) for input_name in inputs)
            if len(bbox) == 1:
                bbox = bbox[0]

            return bbox

    def __eq__(self, value):
        """Note equality can be either with old representation or new one."""
        if isinstance(value, tuple):
            return self.bounding_box() == value
        elif isinstance(value, ModelBoundingBox):
            return (self.intervals == value.intervals) and (
                self.ignored == value.ignored
            )
        else:
            return False

    def __setitem__(self, key, value):
        """Validate and store interval under key (input index or input name)."""
        index = self._get_index(key)
        if index in self._ignored:
            self._ignored.remove(index)

        self._intervals[index] = _Interval.validate(value)

    def __delitem__(self, key):
        """Delete stored interval."""
        index = self._get_index(key)
        if index in self._ignored:
            raise RuntimeError(f"Cannot delete ignored input: {key}!")
        del self._intervals[index]
        self._ignored.append(index)

    def _validate_dict(self, bounding_box: dict):
        """Validate passing dictionary of intervals and setting them."""
        for key, value in bounding_box.items():
            self[key] = value

    @property
    def _available_input_index(self):
        model_input_index = [self._get_index(_input) for _input in self._model.inputs]

        return [_input for _input in model_input_index if _input not in self._ignored]

    def _validate_sequence(self, bounding_box, order: str | None = None):
        """
        Validate passing tuple of tuples representation (or related) and setting them.
        """
        order = self._get_order(order)
        if order == "C":
            # If bounding_box is C/python ordered, it needs to be reversed
            # to be in Fortran/mathematical/input order.
            bounding_box = bounding_box[::-1]

        for index, value in enumerate(bounding_box):
            self[self._available_input_index[index]] = value

    @property
    def _n_inputs(self) -> int:
        n_inputs = self._model.n_inputs - len(self._ignored)
        if n_inputs > 0:
            return n_inputs
        else:
            return 0

    def _validate_iterable(self, bounding_box, order: str | None = None):
        """Validate and set any iterable representation."""
        if len(bounding_box) != self._n_inputs:
            raise ValueError(
                f"Found {len(bounding_box)} intervals, "
                f"but must have exactly {self._n_inputs}."
            )

        if isinstance(bounding_box, dict):
            self._validate_dict(bounding_box)
        else:
            self._validate_sequence(bounding_box, order)

    def _validate(self, bounding_box, order: str | None = None):
        """Validate and set any representation."""
        if self._n_inputs == 1 and not isinstance(bounding_box, dict):
            self[self._available_input_index[0]] = bounding_box
        else:
            self._validate_iterable(bounding_box, order)

    @classmethod
    def validate(
        cls,
        model,
        bounding_box,
        ignored: list | None = None,
        order: str = "C",
        _preserve_ignore: bool = False,
        **kwargs,
    ) -> Self:
        """
        Construct a valid bounding box for a model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The model for which this will be a bounding_box
        bounding_box : dict, tuple
            A possible representation of the bounding box
        order : optional, str
            The order that a tuple representation will be assumed to be
                Default: 'C'
        """
        if isinstance(bounding_box, ModelBoundingBox):
            order = bounding_box.order
            if _preserve_ignore:
                ignored = bounding_box.ignored
            bounding_box = bounding_box.named_intervals

        new = cls({}, model, ignored=ignored, order=order)
        new._validate(bounding_box)

        return new

    def fix_inputs(self, model, fixed_inputs: dict, _keep_ignored=False) -> Self:
        """
        Fix the bounding_box for a `fix_inputs` compound model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The new model for which this will be a bounding_box
        fixed_inputs : dict
            Dictionary of inputs which have been fixed by this bounding box.
        keep_ignored : bool
            Keep the ignored inputs of the bounding box (internal argument only)
        """
        new = self.copy()

        for _input in fixed_inputs.keys():
            del new[_input]

        if _keep_ignored:
            ignored = new.ignored
        else:
            ignored = None

        return ModelBoundingBox.validate(
            model, new.named_intervals, ignored=ignored, order=new._order
        )

    @property
    def dimension(self):
        return len(self)

    def domain(self, resolution, order: str | None = None) -> list[np.ndarray]:
        inputs = self._model.inputs
        order = self._get_order(order)
        if order == "C":
            inputs = inputs[::-1]

        return [self[input_name].domain(resolution) for input_name in inputs]

    def _outside(self, input_shape, inputs):
        """
        Get all the input positions which are outside the bounding_box,
        so that the corresponding outputs can be filled with the fill
        value (default NaN).

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        inputs : list
            List of all the model inputs

        Returns
        -------
        outside_index : bool-numpy array
            True  -> position outside bounding_box
            False -> position inside  bounding_box
        all_out : bool
            if all of the inputs are outside the bounding_box
        """
        all_out = False

        outside_index = np.zeros(input_shape, dtype=bool)
        for index, _input in enumerate(inputs):
            _input = np.asanyarray(_input)

            outside = np.broadcast_to(self[index].outside(_input), input_shape)
            outside_index[outside] = True

            if outside_index.all():
                all_out = True
                break

        return outside_index, all_out

    def _valid_index(self, input_shape, inputs):
        """
        Get the indices of all the inputs inside the bounding_box.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        inputs : list
            List of all the model inputs

        Returns
        -------
        valid_index : numpy array
            array of all indices inside the bounding box
        all_out : bool
            if all of the inputs are outside the bounding_box
        """
        outside_index, all_out = self._outside(input_shape, inputs)

        valid_index = np.atleast_1d(np.logical_not(outside_index)).nonzero()
        if len(valid_index[0]) == 0:
            all_out = True

        return valid_index, all_out

    def prepare_inputs(self, input_shape, inputs) -> tuple[Any, Any, Any]:
        """
        Get prepare the inputs with respect to the bounding box.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        inputs : list
            List of all the model inputs

        Returns
        -------
        valid_inputs : list
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : array_like
            array of all indices inside the bounding box
        all_out: bool
            if all of the inputs are outside the bounding_box
        """
        valid_index, all_out = self._valid_index(input_shape, inputs)

        valid_inputs = []
        if not all_out:
            for _input in inputs:
                if input_shape:
                    valid_input = np.broadcast_to(np.atleast_1d(_input), input_shape)[
                        valid_index
                    ]
                    if np.isscalar(_input):
                        valid_input = valid_input.item(0)
                    valid_inputs.append(valid_input)
                else:
                    valid_inputs.append(_input)

        return tuple(valid_inputs), valid_index, all_out


class _BaseSelectorArgument(NamedTuple):
    index: int
    ignore: bool


class _SelectorArgument(_BaseSelectorArgument):
    """
    Contains a single CompoundBoundingBox slicing input.

    Parameters
    ----------
    index : int
        The index of the input in the input list

    ignore : bool
        Whether or not this input will be ignored by the bounding box.

    Methods
    -------
    validate :
        Returns a valid SelectorArgument for a given model.

    get_selector :
        Returns the value of the input for use in finding the correct
        bounding_box.

    get_fixed_value :
        Gets the slicing value from a fix_inputs set of values.
    """

    def __new__(cls, index, ignore):
        self = super().__new__(cls, index, ignore)

        return self

    @classmethod
    def validate(cls, model, argument, ignored: bool = True) -> Self:
        """
        Construct a valid selector argument for a CompoundBoundingBox.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The model for which this will be an argument for.
        argument : int or str
            A representation of which evaluation input to use
        ignored : optional, bool
            Whether or not to ignore this argument in the ModelBoundingBox.

        Returns
        -------
        Validated selector_argument
        """
        return cls(get_index(model, argument), ignored)

    def get_selector(self, *inputs):
        """
        Get the selector value corresponding to this argument.

        Parameters
        ----------
        *inputs :
            All the processed model evaluation inputs.
        """
        _selector = inputs[self.index]
        if np.iterable(_selector):
            if len(_selector) == 1:
                return _selector[0]
            else:
                return tuple(_selector)
        return _selector

    def name(self, model) -> str:
        """
        Get the name of the input described by this selector argument.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model this selector argument is for.
        """
        return get_name(model, self.index)

    def pretty_repr(self, model):
        """
        Get a pretty-print representation of this object.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model this selector argument is for.
        """
        return f"Argument(name='{self.name(model)}', ignore={self.ignore})"

    def get_fixed_value(self, model, values: dict):
        """
        Gets the value fixed input corresponding to this argument.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model this selector argument is for.

        values : dict
            Dictionary of fixed inputs.
        """
        if self.index in values:
            return values[self.index]
        else:
            if self.name(model) in values:
                return values[self.name(model)]
            else:
                raise RuntimeError(
                    f"{self.pretty_repr(model)} was not found in {values}"
                )

    def is_argument(self, model, argument) -> bool:
        """
        Determine if passed argument is described by this selector argument.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model this selector argument is for.

        argument : int or str
            A representation of which evaluation input is being used
        """
        return self.index == get_index(model, argument)

    def named_tuple(self, model):
        """
        Get a tuple representation of this argument using the input
        name from the model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model this selector argument is for.
        """
        return (self.name(model), self.ignore)


class _SelectorArguments(tuple):
    """
    Contains the CompoundBoundingBox slicing description.

    Parameters
    ----------
    input_ :
        The SelectorArgument values

    Methods
    -------
    validate :
        Returns a valid SelectorArguments for its model.

    get_selector :
        Returns the selector a set of inputs corresponds to.

    is_selector :
        Determines if a selector is correctly formatted for this CompoundBoundingBox.

    get_fixed_value :
        Gets the selector from a fix_inputs set of values.
    """

    _kept_ignore = None

    def __new__(
        cls, input_: tuple[_SelectorArgument], kept_ignore: list | None = None
    ) -> Self:
        self = super().__new__(cls, input_)

        if kept_ignore is None:
            self._kept_ignore = []
        else:
            self._kept_ignore = kept_ignore

        return self

    def pretty_repr(self, model):
        """
        Get a pretty-print representation of this object.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.
        """
        parts = ["SelectorArguments("]
        for argument in self:
            parts.append(f"    {argument.pretty_repr(model)}")
        parts.append(")")

        return "\n".join(parts)

    @property
    def ignore(self):
        """Get the list of ignored inputs."""
        ignore = [argument.index for argument in self if argument.ignore]
        ignore.extend(self._kept_ignore)

        return ignore

    @property
    def kept_ignore(self):
        """The arguments to persist in ignoring."""
        return self._kept_ignore

    @classmethod
    def validate(cls, model, arguments, kept_ignore: list | None = None) -> Self:
        """
        Construct a valid Selector description for a CompoundBoundingBox.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        arguments :
            The individual argument information

        kept_ignore :
            Arguments to persist as ignored
        """
        inputs = []
        for argument in arguments:
            _input = _SelectorArgument.validate(model, *argument)
            if _input.index in [this.index for this in inputs]:
                raise ValueError(
                    f"Input: '{get_name(model, _input.index)}' has been repeated."
                )
            inputs.append(_input)

        if len(inputs) == 0:
            raise ValueError("There must be at least one selector argument.")

        if isinstance(arguments, _SelectorArguments):
            if kept_ignore is None:
                kept_ignore = []

            kept_ignore.extend(arguments.kept_ignore)

        return cls(tuple(inputs), kept_ignore)

    def get_selector(self, *inputs):
        """
        Get the selector corresponding to these inputs.

        Parameters
        ----------
        *inputs :
            All the processed model evaluation inputs.
        """
        return tuple(argument.get_selector(*inputs) for argument in self)

    def is_selector(self, _selector):
        """
        Determine if this is a reasonable selector.

        Parameters
        ----------
        _selector : tuple
            The selector to check
        """
        return isinstance(_selector, tuple) and len(_selector) == len(self)

    def get_fixed_values(self, model, values: dict):
        """
        Gets the value fixed input corresponding to this argument.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        values : dict
            Dictionary of fixed inputs.
        """
        return tuple(argument.get_fixed_value(model, values) for argument in self)

    def is_argument(self, model, argument) -> bool:
        """
        Determine if passed argument is one of the selector arguments.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        argument : int or str
            A representation of which evaluation input is being used
        """
        return any(selector_arg.is_argument(model, argument) for selector_arg in self)

    def selector_index(self, model, argument):
        """
        Get the index of the argument passed in the selector tuples.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        argument : int or str
            A representation of which argument is being used
        """
        for index, selector_arg in enumerate(self):
            if selector_arg.is_argument(model, argument):
                return index
        raise ValueError(f"{argument} does not correspond to any selector argument.")

    def reduce(self, model, argument):
        """
        Reduce the selector arguments by the argument given.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        argument : int or str
            A representation of which argument is being used
        """
        arguments = list(self)
        kept_ignore = [arguments.pop(self.selector_index(model, argument)).index]
        kept_ignore.extend(self._kept_ignore)

        return _SelectorArguments.validate(model, tuple(arguments), kept_ignore)

    def add_ignore(self, model, argument):
        """
        Add argument to the kept_ignore list.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.

        argument : int or str
            A representation of which argument is being used
        """
        if self.is_argument(model, argument):
            raise ValueError(
                f"{argument}: is a selector argument and cannot be ignored."
            )

        kept_ignore = [get_index(model, argument)]

        return _SelectorArguments.validate(model, self, kept_ignore)

    def named_tuple(self, model):
        """
        Get a tuple of selector argument tuples using input names.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The Model these selector arguments are for.
        """
        return tuple(selector_arg.named_tuple(model) for selector_arg in self)


class CompoundBoundingBox(_BoundingDomain):
    """
    A model's compound bounding box.

    Parameters
    ----------
    bounding_boxes : dict
        A dictionary containing all the ModelBoundingBoxes that are possible
            keys   -> _selector (extracted from model inputs)
            values -> ModelBoundingBox

    model : `~astropy.modeling.Model`
        The Model this compound bounding_box is for.

    selector_args : _SelectorArguments
        A description of how to extract the selectors from model inputs.

    create_selector : optional
        A method which takes in the selector and the model to return a
        valid bounding corresponding to that selector. This can be used
        to construct new bounding_boxes for previously undefined selectors.
        These new boxes are then stored for future lookups.

    order : optional, str
        The ordering that is assumed for the tuple representation of the
        bounding_boxes.
    """

    def __init__(
        self,
        bounding_boxes: dict[Any, ModelBoundingBox],
        model,
        selector_args: _SelectorArguments,
        create_selector: Callable | None = None,
        ignored: list[int] | None = None,
        order: str = "C",
    ):
        super().__init__(model, ignored, order)

        self._create_selector = create_selector
        self._selector_args = _SelectorArguments.validate(model, selector_args)

        self._bounding_boxes = {}
        self._validate(bounding_boxes)

    def copy(self):
        bounding_boxes = {
            selector: bbox.copy(self.selector_args.ignore)
            for selector, bbox in self._bounding_boxes.items()
        }

        return CompoundBoundingBox(
            bounding_boxes,
            self._model,
            selector_args=self._selector_args,
            create_selector=copy.deepcopy(self._create_selector),
            order=self._order,
        )

    def __repr__(self):
        parts = ["CompoundBoundingBox(", "    bounding_boxes={"]
        # bounding_boxes
        for _selector, bbox in self._bounding_boxes.items():
            bbox_repr = bbox.__repr__().split("\n")
            parts.append(f"        {_selector} = {bbox_repr.pop(0)}")
            for part in bbox_repr:
                parts.append(f"            {part}")
        parts.append("    }")

        # selector_args
        selector_args_repr = self.selector_args.pretty_repr(self._model).split("\n")
        parts.append(f"    selector_args = {selector_args_repr.pop(0)}")
        for part in selector_args_repr:
            parts.append(f"        {part}")
        parts.append(")")

        return "\n".join(parts)

    @property
    def bounding_boxes(self) -> dict[Any, ModelBoundingBox]:
        return self._bounding_boxes

    @property
    def selector_args(self) -> _SelectorArguments:
        return self._selector_args

    @selector_args.setter
    def selector_args(self, value):
        self._selector_args = _SelectorArguments.validate(self._model, value)

        warnings.warn(
            "Overriding selector_args may cause problems you should re-validate "
            "the compound bounding box before use!",
            RuntimeWarning,
        )

    @property
    def named_selector_tuple(self) -> tuple:
        return self._selector_args.named_tuple(self._model)

    @property
    def create_selector(self):
        return self._create_selector

    @staticmethod
    def _get_selector_key(key):
        if np.iterable(key):
            return tuple(key)
        else:
            return (key,)

    def __setitem__(self, key, value):
        _selector = self._get_selector_key(key)
        if not self.selector_args.is_selector(_selector):
            raise ValueError(f"{_selector} is not a selector!")

        ignored = self.selector_args.ignore + self.ignored
        self._bounding_boxes[_selector] = ModelBoundingBox.validate(
            self._model, value, ignored, order=self._order
        )

    def _validate(self, bounding_boxes: dict):
        for _selector, bounding_box in bounding_boxes.items():
            self[_selector] = bounding_box

    def __eq__(self, value):
        if isinstance(value, CompoundBoundingBox):
            return (
                self.bounding_boxes == value.bounding_boxes
                and self.selector_args == value.selector_args
                and self.create_selector == value.create_selector
            )
        else:
            return False

    @classmethod
    def validate(
        cls,
        model,
        bounding_box: dict,
        selector_args=None,
        create_selector=None,
        ignored: list | None = None,
        order: str = "C",
        _preserve_ignore: bool = False,
        **kwarg,
    ) -> Self:
        """
        Construct a valid compound bounding box for a model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The model for which this will be a bounding_box
        bounding_box : dict
            Dictionary of possible bounding_box representations
        selector_args : optional
            Description of the selector arguments
        create_selector : optional, callable
            Method for generating new selectors
        order : optional, str
            The order that a tuple representation will be assumed to be
                Default: 'C'
        """
        if isinstance(bounding_box, CompoundBoundingBox):
            if selector_args is None:
                selector_args = bounding_box.selector_args
            if create_selector is None:
                create_selector = bounding_box.create_selector
            order = bounding_box.order
            if _preserve_ignore:
                ignored = bounding_box.ignored
            bounding_box = bounding_box.bounding_boxes

        if selector_args is None:
            raise ValueError(
                "Selector arguments must be provided "
                "(can be passed as part of bounding_box argument)"
            )

        return cls(
            bounding_box,
            model,
            selector_args,
            create_selector=create_selector,
            ignored=ignored,
            order=order,
        )

    def __contains__(self, key):
        return key in self._bounding_boxes

    def _create_bounding_box(self, _selector):
        self[_selector] = self._create_selector(_selector, model=self._model)

        return self[_selector]

    def __getitem__(self, key):
        _selector = self._get_selector_key(key)
        if _selector in self:
            return self._bounding_boxes[_selector]
        elif self._create_selector is not None:
            return self._create_bounding_box(_selector)
        else:
            raise RuntimeError(f"No bounding box is defined for selector: {_selector}.")

    def _select_bounding_box(self, inputs) -> ModelBoundingBox:
        _selector = self.selector_args.get_selector(*inputs)

        return self[_selector]

    def prepare_inputs(self, input_shape, inputs) -> tuple[Any, Any, Any]:
        """
        Get prepare the inputs with respect to the bounding box.

        Parameters
        ----------
        input_shape : tuple
            The shape that all inputs have be reshaped/broadcasted into
        inputs : list
            List of all the model inputs

        Returns
        -------
        valid_inputs : list
            The inputs reduced to just those inputs which are all inside
            their respective bounding box intervals
        valid_index : array_like
            array of all indices inside the bounding box
        all_out: bool
            if all of the inputs are outside the bounding_box
        """
        bounding_box = self._select_bounding_box(inputs)
        return bounding_box.prepare_inputs(input_shape, inputs)

    def _matching_bounding_boxes(self, argument, value) -> dict[Any, ModelBoundingBox]:
        selector_index = self.selector_args.selector_index(self._model, argument)
        matching = {}
        for selector_key, bbox in self._bounding_boxes.items():
            if selector_key[selector_index] == value:
                new_selector_key = list(selector_key)
                new_selector_key.pop(selector_index)

                if bbox.has_interval(argument):
                    new_bbox = bbox.fix_inputs(
                        self._model, {argument: value}, _keep_ignored=True
                    )
                else:
                    new_bbox = bbox.copy()

                matching[tuple(new_selector_key)] = new_bbox

        if len(matching) == 0:
            raise ValueError(
                f"Attempting to fix input {argument}, but there are no "
                f"bounding boxes for argument value {value}."
            )

        return matching

    def _fix_input_selector_arg(self, argument, value):
        matching_bounding_boxes = self._matching_bounding_boxes(argument, value)

        if len(self.selector_args) == 1:
            return matching_bounding_boxes[()]
        else:
            return CompoundBoundingBox(
                matching_bounding_boxes,
                self._model,
                self.selector_args.reduce(self._model, argument),
            )

    def _fix_input_bbox_arg(self, argument, value):
        bounding_boxes = {}
        for selector_key, bbox in self._bounding_boxes.items():
            bounding_boxes[selector_key] = bbox.fix_inputs(
                self._model, {argument: value}, _keep_ignored=True
            )

        return CompoundBoundingBox(
            bounding_boxes,
            self._model,
            self.selector_args.add_ignore(self._model, argument),
        )

    def fix_inputs(self, model, fixed_inputs: dict) -> Self:
        """
        Fix the bounding_box for a `fix_inputs` compound model.

        Parameters
        ----------
        model : `~astropy.modeling.Model`
            The new model for which this will be a bounding_box
        fixed_inputs : dict
            Dictionary of inputs which have been fixed by this bounding box.
        """
        fixed_input_keys = list(fixed_inputs.keys())
        argument = fixed_input_keys.pop()
        value = fixed_inputs[argument]

        if self.selector_args.is_argument(self._model, argument):
            bbox = self._fix_input_selector_arg(argument, value)
        else:
            bbox = self._fix_input_bbox_arg(argument, value)

        if len(fixed_input_keys) > 0:
            new_fixed_inputs = fixed_inputs.copy()
            del new_fixed_inputs[argument]

            bbox = bbox.fix_inputs(model, new_fixed_inputs)

        if isinstance(bbox, CompoundBoundingBox):
            selector_args = bbox.named_selector_tuple
            bbox_dict = bbox
        elif isinstance(bbox, ModelBoundingBox):
            selector_args = None
            bbox_dict = bbox.named_intervals

        return bbox.__class__.validate(
            model, bbox_dict, order=bbox.order, selector_args=selector_args
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import sys
from os.path import relpath
from pathlib import Path

from setuptools import Extension

ASTROPY_COSMOLOGY_SRC_ROOT = Path(__file__).parent


if sys.platform.startswith("win"):
    # on windows, -Werror (and possibly -Wall too) isn't recognized
    extra_compile_args = []
else:
    # be extra careful with this extension as it calls PyErr_WarnEx
    # with a formatted message whose size cannot be determined at compile time,
    # which is never done within the standard library
    extra_compile_args = ["-Werror", "-Wall"]


def get_extensions():
    return [
        Extension(
            "astropy.cosmology._src.signature_deprecations",
            [relpath(Path(ASTROPY_COSMOLOGY_SRC_ROOT, "signature_deprecations.c"))],
            extra_compile_args=extra_compile_args,
        ),
    ]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Accuracy tests for ICRS transformations, primarily to/from AltAz."""

import numpy as np

from astropy import units as u
from astropy.coordinates import (
    CIRS,
    ICRS,
    AltAz,
    EarthLocation,
    HADec,
    SkyCoord,
    frame_transform_graph,
    golden_spiral_grid,
)
from astropy.tests.helper import assert_quantity_allclose as assert_allclose
from astropy.time import Time


def test_icrs_altaz_consistency():
    """
    Check ICRS<->AltAz for consistency with ICRS<->CIRS<->AltAz

    The latter is extensively tested in test_intermediate_transformations.py
    """
    usph = golden_spiral_grid(200)
    dist = np.linspace(0.5, 1, len(usph)) * u.km * 1e5
    icoo = SkyCoord(ra=usph.lon, dec=usph.lat, distance=dist)

    observer = EarthLocation(28 * u.deg, 23 * u.deg, height=2000.0 * u.km)
    obstime = Time("J2010")
    aa_frame = AltAz(obstime=obstime, location=observer)

    # check we are going direct!
    trans = frame_transform_graph.get_transform(ICRS, AltAz).transforms
    assert len(trans) == 1

    # check that ICRS-AltAz and ICRS->CIRS->AltAz are consistent
    aa1 = icoo.transform_to(aa_frame)
    aa2 = icoo.transform_to(CIRS()).transform_to(aa_frame)
    assert_allclose(aa1.separation_3d(aa2), 0 * u.mm, atol=1 * u.mm)

    # check roundtrip
    roundtrip = icoo.transform_to(aa_frame).transform_to(icoo)
    assert_allclose(roundtrip.separation_3d(icoo), 0 * u.mm, atol=1 * u.mm)

    # check there and back via CIRS mish-mash
    roundtrip = icoo.transform_to(aa_frame).transform_to(CIRS()).transform_to(icoo)
    assert_allclose(roundtrip.separation_3d(icoo), 0 * u.mm, atol=1 * u.mm)


def test_icrs_hadec_consistency():
    """
    Check ICRS<->HADec for consistency with ICRS<->CIRS<->HADec
    """
    usph = golden_spiral_grid(200)
    dist = np.linspace(0.5, 1, len(usph)) * u.km * 1e5
    icoo = SkyCoord(ra=usph.lon, dec=usph.lat, distance=dist)

    observer = EarthLocation(28 * u.deg, 23 * u.deg, height=2000.0 * u.km)
    obstime = Time("J2010")
    hd_frame = HADec(obstime=obstime, location=observer)

    # check we are going direct!
    trans = frame_transform_graph.get_transform(ICRS, HADec).transforms
    assert len(trans) == 1

    # check that ICRS-HADec and ICRS->CIRS->HADec are consistent
    aa1 = icoo.transform_to(hd_frame)
    aa2 = icoo.transform_to(CIRS()).transform_to(hd_frame)
    assert_allclose(aa1.separation_3d(aa2), 0 * u.mm, atol=1 * u.mm)

    # check roundtrip
    roundtrip = icoo.transform_to(hd_frame).transform_to(icoo)
    assert_allclose(roundtrip.separation_3d(icoo), 0 * u.mm, atol=1 * u.mm)

    # check there and back via CIRS mish-mash
    roundtrip = icoo.transform_to(hd_frame).transform_to(CIRS()).transform_to(icoo)
    assert_allclose(roundtrip.separation_3d(icoo), 0 * u.mm, atol=1 * u.mm)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest

from astropy.io import fits
from astropy.io.fits.column import Column
from astropy.io.fits.diff import (
    FITSDiff,
    HDUDiff,
    HeaderDiff,
    ImageDataDiff,
    TableDataDiff,
)
from astropy.io.fits.hdu import HDUList, ImageHDU, PrimaryHDU
from astropy.io.fits.hdu.base import NonstandardExtHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.header import Header
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

from .conftest import FitsTestCase


class DummyNonstandardExtHDU(NonstandardExtHDU):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self._buffer = np.asarray(data).tobytes()
        self._data_offset = 0

    @property
    def size(self):
        return len(self._buffer)


class TestDiff(FitsTestCase):
    def test_identical_headers(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        assert HeaderDiff(ha, hb).identical
        assert HeaderDiff(ha.tostring(), hb.tostring()).identical

        with pytest.raises(TypeError):
            HeaderDiff(1, 2)

    def test_slightly_different_headers(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        assert not HeaderDiff(ha, hb).identical

    def test_common_keywords(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        hb["D"] = (5, "Comment")
        assert HeaderDiff(ha, hb).common_keywords == ["A", "B", "C"]

    def test_different_keyword_count(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        del hb["B"]
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_count == (3, 2)

        # But make sure the common keywords are at least correct
        assert diff.common_keywords == ["A", "C"]

    def test_different_keywords(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        hb["D"] = (5, "Comment")
        ha["E"] = (6, "Comment")
        ha["F"] = (7, "Comment")
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keywords == (["E", "F"], ["D"])

    def test_different_keyword_values(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_values == {"C": [(3, 4)]}

    def test_different_keyword_comments(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3, "comment 1")])
        hb = ha.copy()
        hb.comments["C"] = "comment 2"
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_comments == {"C": [("comment 1", "comment 2")]}

    def test_different_keyword_values_with_duplicate(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        ha.append(("C", 4))
        hb.append(("C", 5))
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_values == {"C": [None, (4, 5)]}

    def test_asymmetric_duplicate_keywords(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        ha.append(("A", 2, "comment 1"))
        ha.append(("A", 3, "comment 2"))
        hb.append(("B", 4, "comment 3"))
        hb.append(("C", 5, "comment 4"))
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_values == {}
        assert diff.diff_duplicate_keywords == {"A": (3, 1), "B": (1, 2), "C": (1, 2)}

        report = diff.report()
        assert (
            "Inconsistent duplicates of keyword 'A'     :\n"
            "  Occurs 3 time(s) in a, 1 times in (b)" in report
        )

    def test_floating_point_rtol(self):
        ha = Header([("A", 1), ("B", 2.00001), ("C", 3.000001)])
        hb = ha.copy()
        hb["B"] = 2.00002
        hb["C"] = 3.000002
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.diff_keyword_values == {
            "B": [(2.00001, 2.00002)],
            "C": [(3.000001, 3.000002)],
        }
        diff = HeaderDiff(ha, hb, rtol=1e-6)
        assert not diff.identical
        assert diff.diff_keyword_values == {"B": [(2.00001, 2.00002)]}
        diff = HeaderDiff(ha, hb, rtol=1e-5)
        assert diff.identical

    def test_floating_point_atol(self):
        ha = Header([("A", 1), ("B", 1.0), ("C", 0.0)])
        hb = ha.copy()
        hb["B"] = 1.00001
        hb["C"] = 0.000001
        diff = HeaderDiff(ha, hb, rtol=1e-6)
        assert not diff.identical
        assert diff.diff_keyword_values == {
            "B": [(1.0, 1.00001)],
            "C": [(0.0, 0.000001)],
        }
        diff = HeaderDiff(ha, hb, rtol=1e-5)
        assert not diff.identical
        assert diff.diff_keyword_values == {"C": [(0.0, 0.000001)]}
        diff = HeaderDiff(ha, hb, atol=1e-6)
        assert not diff.identical
        assert diff.diff_keyword_values == {"B": [(1.0, 1.00001)]}
        diff = HeaderDiff(ha, hb, atol=1e-5)  # strict inequality
        assert not diff.identical
        assert diff.diff_keyword_values == {"B": [(1.0, 1.00001)]}
        diff = HeaderDiff(ha, hb, rtol=1e-5, atol=1e-5)
        assert diff.identical
        diff = HeaderDiff(ha, hb, atol=1.1e-5)
        assert diff.identical
        diff = HeaderDiff(ha, hb, rtol=1e-6, atol=1e-6)
        assert not diff.identical

    def test_ignore_blanks(self):
        with fits.conf.set_temp("strip_header_whitespace", False):
            ha = Header([("A", 1), ("B", 2), ("C", "A       ")])
            hb = ha.copy()
            hb["C"] = "A"
            assert ha["C"] != hb["C"]

            diff = HeaderDiff(ha, hb)
            # Trailing blanks are ignored by default
            assert diff.identical
            assert diff.diff_keyword_values == {}

            # Don't ignore blanks
            diff = HeaderDiff(ha, hb, ignore_blanks=False)
            assert not diff.identical
            assert diff.diff_keyword_values == {"C": [("A       ", "A")]}

    @pytest.mark.parametrize("differ", [HeaderDiff, HDUDiff, FITSDiff])
    def test_ignore_blank_cards(self, differ):
        """Test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/152

        Ignore blank cards.
        """

        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = Header([("A", 1), ("", ""), ("B", 2), ("", ""), ("C", 3)])
        hc = ha.copy()
        if differ is HeaderDiff:
            hc.append()
            hc.append()
        else:  # Ensure blanks are not at the end as they are stripped by HDUs
            hc.add_blank(after=-2)
            hc.add_blank(after=-2)

        if differ in (HDUDiff, FITSDiff):  # wrap it in a PrimaryHDU
            ha, hb, hc = (PrimaryHDU(np.arange(10), h) for h in (ha, hb, hc))
            hc_header = hc.header
        if differ is FITSDiff:  # wrap it in a HDUList
            ha, hb, hc = (HDUList([h]) for h in (ha, hb, hc))
            hc_header = hc[0].header

        # We now have a header with interleaved blanks, and a header with end
        # blanks, both of which should ignore the blanks
        assert differ(ha, hb).identical
        assert differ(ha, hc).identical
        assert differ(hb, hc).identical

        assert not differ(ha, hb, ignore_blank_cards=False).identical
        assert not differ(ha, hc, ignore_blank_cards=False).identical

        # Both hb and hc have the same number of blank cards; since order is
        # currently ignored, these should still be identical even if blank
        # cards are not ignored
        assert differ(hb, hc, ignore_blank_cards=False).identical

        if differ is HeaderDiff:
            hc.append()
        else:  # Ensure blanks are not at the end as they are stripped by HDUs
            hc_header.add_blank(after=-2)
        # But now there are different numbers of blanks, so they should not be
        # ignored:
        assert not differ(hb, hc, ignore_blank_cards=False).identical

    def test_ignore_hdus(self):
        a = np.arange(100).reshape(10, 10)
        b = a.copy()
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        xa = np.array([(1.0, 1), (3.0, 4)], dtype=[("x", float), ("y", int)])
        xb = np.array([(1.0, 2), (3.0, 5)], dtype=[("x", float), ("y", int)])
        phdu = PrimaryHDU(header=ha)
        ihdua = ImageHDU(data=a, name="SCI")
        ihdub = ImageHDU(data=b, name="SCI")
        bhdu1 = BinTableHDU(data=xa, name="ASDF")
        bhdu2 = BinTableHDU(data=xb, name="ASDF")
        hdula = HDUList([phdu, ihdua, bhdu1])
        hdulb = HDUList([phdu, ihdub, bhdu2])

        # ASDF extension should be different
        diff = FITSDiff(hdula, hdulb)
        assert not diff.identical
        assert diff.diff_hdus[0][0] == 2

        # ASDF extension should be ignored
        diff = FITSDiff(hdula, hdulb, ignore_hdus=["ASDF"])
        assert diff.identical, diff.report()

        diff = FITSDiff(hdula, hdulb, ignore_hdus=["ASD*"])
        assert diff.identical, diff.report()

        # SCI extension should be different
        hdulb["SCI"].data += 1
        diff = FITSDiff(hdula, hdulb, ignore_hdus=["ASDF"])
        assert not diff.identical

        # SCI and ASDF extensions should be ignored
        diff = FITSDiff(hdula, hdulb, ignore_hdus=["SCI", "ASDF"])
        assert diff.identical, diff.report()

        # All EXTVER of SCI should be ignored
        ihduc = ImageHDU(data=a, name="SCI", ver=2)
        hdulb.append(ihduc)
        diff = FITSDiff(hdula, hdulb, ignore_hdus=["SCI", "ASDF"])
        assert not any(diff.diff_hdus), diff.report()
        assert any(diff.diff_hdu_count), diff.report()

    def test_ignore_keyword_values(self):
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["B"] = 4
        hb["C"] = 5
        diff = HeaderDiff(ha, hb, ignore_keywords=["*"])
        assert diff.identical
        diff = HeaderDiff(ha, hb, ignore_keywords=["B"])
        assert not diff.identical
        assert diff.diff_keyword_values == {"C": [(3, 5)]}

        report = diff.report()
        assert "Keyword B        has different values" not in report
        assert "Keyword C        has different values" in report

        # Test case-insensitivity
        diff = HeaderDiff(ha, hb, ignore_keywords=["b"])
        assert not diff.identical
        assert diff.diff_keyword_values == {"C": [(3, 5)]}

    def test_ignore_keyword_comments(self):
        ha = Header([("A", 1, "A"), ("B", 2, "B"), ("C", 3, "C")])
        hb = ha.copy()
        hb.comments["B"] = "D"
        hb.comments["C"] = "E"
        diff = HeaderDiff(ha, hb, ignore_comments=["*"])
        assert diff.identical
        diff = HeaderDiff(ha, hb, ignore_comments=["B"])
        assert not diff.identical
        assert diff.diff_keyword_comments == {"C": [("C", "E")]}

        report = diff.report()
        assert "Keyword B        has different comments" not in report
        assert "Keyword C        has different comments" in report

        # Test case-insensitivity
        diff = HeaderDiff(ha, hb, ignore_comments=["b"])
        assert not diff.identical
        assert diff.diff_keyword_comments == {"C": [("C", "E")]}

    def test_hierarch_keywords_identical(self):
        ha = Header(
            [
                ("HIERARCH UPPER", 1),
                ("HIERARCH lower", 2),
                ("HIERARCH veryverylong", 3),
            ]
        )
        hb = ha.copy()
        assert HeaderDiff(ha, hb).identical

    def test_hierarch_keywords_different(self):
        ha = Header([("HIERARCH Both", 1)])
        hb = Header([("HIERARCH BOTh", 1)])
        diff = HeaderDiff(ha, hb)
        assert not diff.identical
        assert diff.common_keywords == []
        assert diff.diff_keywords == (["Both"], ["BOTh"])

    def test_trivial_identical_images(self):
        ia = np.arange(100).reshape(10, 10)
        ib = np.arange(100).reshape(10, 10)
        diff = ImageDataDiff(ia, ib)
        assert diff.identical
        assert diff.diff_total == 0

    def test_identical_within_relative_tolerance(self):
        ia = np.ones((10, 10)) - 0.00001
        ib = np.ones((10, 10)) - 0.00002
        diff = ImageDataDiff(ia, ib, rtol=1.0e-4)
        assert diff.identical
        assert diff.diff_total == 0

    def test_identical_within_absolute_tolerance(self):
        ia = np.zeros((10, 10)) - 0.00001
        ib = np.zeros((10, 10)) - 0.00002
        diff = ImageDataDiff(ia, ib, rtol=1.0e-4)
        assert not diff.identical
        assert diff.diff_total == 100
        diff = ImageDataDiff(ia, ib, atol=1.0e-4)
        assert diff.identical
        assert diff.diff_total == 0

    def test_identical_within_rtol_and_atol(self):
        ia = np.zeros((10, 10)) - 0.00001
        ib = np.zeros((10, 10)) - 0.00002
        diff = ImageDataDiff(ia, ib, rtol=1.0e-5, atol=1.0e-5)
        assert diff.identical
        assert diff.diff_total == 0

    def test_not_identical_within_rtol_and_atol(self):
        ia = np.zeros((10, 10)) - 0.00001
        ib = np.zeros((10, 10)) - 0.00002
        diff = ImageDataDiff(ia, ib, rtol=1.0e-5, atol=1.0e-6)
        assert not diff.identical
        assert diff.diff_total == 100

    def test_identical_comp_image_hdus(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/189

        For this test we mostly just care that comparing to compressed images
        does not crash, and returns the correct results.  Two compressed images
        will be considered identical if the decompressed data is the same.
        Obviously we test whether or not the same compression was used by
        looking for (or ignoring) header differences.
        """

        data = np.arange(100.0).reshape(10, 10)
        hdu = fits.CompImageHDU(data=data)
        hdu.writeto(self.temp("test.fits"))

        with (
            fits.open(self.temp("test.fits")) as hdula,
            fits.open(self.temp("test.fits")) as hdulb,
        ):
            diff = FITSDiff(hdula, hdulb)
            assert diff.identical

    def test_different_dimensions(self):
        ia = np.arange(100).reshape(10, 10)
        ib = np.arange(100) - 1

        # Although ib could be reshaped into the same dimensions, for now the
        # data is not compared anyways
        diff = ImageDataDiff(ia, ib)
        assert not diff.identical
        assert diff.diff_dimensions == ((10, 10), (100,))
        assert diff.diff_total == 0

        report = diff.report()
        assert "Data dimensions differ" in report
        assert "a: 10 x 10" in report
        assert "b: 100" in report
        assert "No further data comparison performed." in report

    def test_different_pixels(self):
        ia = np.arange(100).reshape(10, 10)
        ib = np.arange(100).reshape(10, 10)
        ib[0, 0] = 10
        ib[5, 5] = 20
        diff = ImageDataDiff(ia, ib)
        assert not diff.identical
        assert diff.diff_dimensions == ()
        assert diff.diff_total == 2
        assert diff.diff_ratio == 0.02
        assert diff.diff_pixels == [((0, 0), (0, 10)), ((5, 5), (55, 20))]

    def test_identical_tables(self):
        c1 = Column("A", format="L", array=[True, False])
        c2 = Column("B", format="X", array=[[0], [1]])
        c3 = Column("C", format="4I", dim="(2, 2)", array=[[0, 1, 2, 3], [4, 5, 6, 7]])
        c4 = Column("D", format="J", bscale=2.0, array=[0, 1])
        c5 = Column("E", format="A3", array=["abc", "def"])
        c6 = Column("F", format="E", unit="m", array=[0.0, 1.0])
        c7 = Column("G", format="D", bzero=-0.1, array=[0.0, 1.0])
        c8 = Column("H", format="C", array=[0.0 + 1.0j, 2.0 + 3.0j])
        c9 = Column("I", format="M", array=[4.0 + 5.0j, 6.0 + 7.0j])
        c10 = Column("J", format="PI(2)", array=[[0, 1], [2, 3]])
        c11 = Column("K", format="QJ(2)", array=[[0, 1], [2, 3]])

        columns = [c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11]

        ta = BinTableHDU.from_columns(columns)
        tb = BinTableHDU.from_columns([c.copy() for c in columns])

        diff = TableDataDiff(ta.data, tb.data)
        assert diff.identical
        assert len(diff.common_columns) == 11
        assert diff.common_column_names == set("abcdefghijk")
        assert diff.diff_ratio == 0
        assert diff.diff_total == 0

    def test_diff_empty_tables(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/178

        Ensure that diffing tables containing empty data doesn't crash.
        """

        c1 = Column("D", format="J")
        c2 = Column("E", format="J")
        thdu = BinTableHDU.from_columns([c1, c2], nrows=0)

        hdula = fits.HDUList([thdu])
        hdulb = fits.HDUList([thdu])

        diff = FITSDiff(hdula, hdulb)
        assert diff.identical

    def test_ignore_table_fields(self):
        c1 = Column("A", format="L", array=[True, False])
        c2 = Column("B", format="X", array=[[0], [1]])
        c3 = Column("C", format="4I", dim="(2, 2)", array=[[0, 1, 2, 3], [4, 5, 6, 7]])

        c4 = Column("B", format="X", array=[[1], [0]])
        c5 = Column("C", format="4I", dim="(2, 2)", array=[[1, 2, 3, 4], [5, 6, 7, 8]])

        ta = BinTableHDU.from_columns([c1, c2, c3])
        tb = BinTableHDU.from_columns([c1, c4, c5])

        diff = TableDataDiff(ta.data, tb.data, ignore_fields=["B", "C"])
        assert diff.identical

        # The only common column should be c1
        assert len(diff.common_columns) == 1
        assert diff.common_column_names == {"a"}
        assert diff.diff_ratio == 0
        assert diff.diff_total == 0

    def test_different_table_field_names(self):
        ca = Column("A", format="L", array=[True, False])
        cb = Column("B", format="L", array=[True, False])
        cc = Column("C", format="L", array=[True, False])

        ta = BinTableHDU.from_columns([ca, cb])
        tb = BinTableHDU.from_columns([ca, cc])

        diff = TableDataDiff(ta.data, tb.data)

        assert not diff.identical
        assert len(diff.common_columns) == 1
        assert diff.common_column_names == {"a"}
        assert diff.diff_column_names == (["B"], ["C"])
        assert diff.diff_ratio == 0
        assert diff.diff_total == 0

        report = diff.report()
        assert "Extra column B of format L in a" in report
        assert "Extra column C of format L in b" in report

    def test_different_table_field_counts(self):
        """
        Test tables with some common columns, but different number of columns
        overall.
        """

        ca = Column("A", format="L", array=[True, False])
        cb = Column("B", format="L", array=[True, False])
        cc = Column("C", format="L", array=[True, False])

        ta = BinTableHDU.from_columns([cb])
        tb = BinTableHDU.from_columns([ca, cb, cc])

        diff = TableDataDiff(ta.data, tb.data)

        assert not diff.identical
        assert diff.diff_column_count == (1, 3)
        assert len(diff.common_columns) == 1
        assert diff.common_column_names == {"b"}
        assert diff.diff_column_names == ([], ["A", "C"])
        assert diff.diff_ratio == 0
        assert diff.diff_total == 0

        report = diff.report()
        assert " Tables have different number of columns:" in report
        assert "  a: 1\n  b: 3" in report

    def test_different_table_rows(self):
        """
        Test tables that are otherwise identical but one has more rows than the
        other.
        """

        ca1 = Column("A", format="L", array=[True, False])
        cb1 = Column("B", format="L", array=[True, False])
        ca2 = Column("A", format="L", array=[True, False, True])
        cb2 = Column("B", format="L", array=[True, False, True])

        ta = BinTableHDU.from_columns([ca1, cb1])
        tb = BinTableHDU.from_columns([ca2, cb2])

        diff = TableDataDiff(ta.data, tb.data)

        assert not diff.identical
        assert diff.diff_column_count == ()
        assert len(diff.common_columns) == 2
        assert diff.diff_rows == (2, 3)
        assert diff.diff_values == []

        report = diff.report()

        assert "Table rows differ" in report
        assert "a: 2" in report
        assert "b: 3" in report
        assert "No further data comparison performed." in report

    def test_different_table_data(self):
        """
        Test diffing table data on columns of several different data formats
        and dimensions.
        """

        ca1 = Column("A", format="L", array=[True, False])
        ca2 = Column("B", format="X", array=[[0], [1]])
        ca3 = Column("C", format="4I", dim="(2, 2)", array=[[0, 1, 2, 3], [4, 5, 6, 7]])
        ca4 = Column("D", format="J", bscale=2.0, array=[0.0, 2.0])
        ca5 = Column("E", format="A3", array=["abc", "def"])
        ca6 = Column("F", format="E", unit="m", array=[0.0, 1.0])
        ca7 = Column("G", format="D", bzero=-0.1, array=[0.0, 1.0])
        ca8 = Column("H", format="C", array=[0.0 + 1.0j, 2.0 + 3.0j])
        ca9 = Column("I", format="M", array=[4.0 + 5.0j, 6.0 + 7.0j])
        ca10 = Column("J", format="PI(2)", array=[[0, 1], [2, 3]])
        ca11 = Column("K", format="QJ(2)", array=[[0, 1], [2, 3]])

        cb1 = Column("A", format="L", array=[False, False])
        cb2 = Column("B", format="X", array=[[0], [0]])
        cb3 = Column("C", format="4I", dim="(2, 2)", array=[[0, 1, 2, 3], [5, 6, 7, 8]])
        cb4 = Column("D", format="J", bscale=2.0, array=[2.0, 2.0])
        cb5 = Column("E", format="A3", array=["abc", "ghi"])
        cb6 = Column("F", format="E", unit="m", array=[1.0, 2.0])
        cb7 = Column("G", format="D", bzero=-0.1, array=[2.0, 3.0])
        cb8 = Column("H", format="C", array=[1.0 + 1.0j, 2.0 + 3.0j])
        cb9 = Column("I", format="M", array=[5.0 + 5.0j, 6.0 + 7.0j])
        cb10 = Column("J", format="PI(2)", array=[[1, 2], [3, 4]])
        cb11 = Column("K", format="QJ(2)", array=[[1, 2], [3, 4]])

        ta = BinTableHDU.from_columns(
            [ca1, ca2, ca3, ca4, ca5, ca6, ca7, ca8, ca9, ca10, ca11]
        )
        tb = BinTableHDU.from_columns(
            [cb1, cb2, cb3, cb4, cb5, cb6, cb7, cb8, cb9, cb10, cb11]
        )

        diff = TableDataDiff(ta.data, tb.data, numdiffs=20)
        assert not diff.identical
        # The column definitions are the same, but not the column values
        assert diff.diff_columns == ()
        assert diff.diff_values[0] == (("A", 0), (True, False))
        assert diff.diff_values[1] == (("B", 1), ([1], [0]))
        assert diff.diff_values[2][0] == ("C", 1)
        assert (diff.diff_values[2][1][0] == [[4, 5], [6, 7]]).all()
        assert (diff.diff_values[2][1][1] == [[5, 6], [7, 8]]).all()
        assert diff.diff_values[3] == (("D", 0), (0, 2.0))
        assert diff.diff_values[4] == (("E", 1), ("def", "ghi"))
        assert diff.diff_values[5] == (("F", 0), (0.0, 1.0))
        assert diff.diff_values[6] == (("F", 1), (1.0, 2.0))
        assert diff.diff_values[7] == (("G", 0), (0.0, 2.0))
        assert diff.diff_values[8] == (("G", 1), (1.0, 3.0))
        assert diff.diff_values[9] == (("H", 0), (0.0 + 1.0j, 1.0 + 1.0j))
        assert diff.diff_values[10] == (("I", 0), (4.0 + 5.0j, 5.0 + 5.0j))
        assert diff.diff_values[11][0] == ("J", 0)
        assert (diff.diff_values[11][1][0] == [0, 1]).all()
        assert (diff.diff_values[11][1][1] == [1, 2]).all()
        assert diff.diff_values[12][0] == ("J", 1)
        assert (diff.diff_values[12][1][0] == [2, 3]).all()
        assert (diff.diff_values[12][1][1] == [3, 4]).all()
        assert diff.diff_values[13][0] == ("K", 0)
        assert (diff.diff_values[13][1][0] == [0, 1]).all()
        assert (diff.diff_values[13][1][1] == [1, 2]).all()
        assert diff.diff_values[14][0] == ("K", 1)
        assert (diff.diff_values[14][1][0] == [2, 3]).all()
        assert (diff.diff_values[14][1][1] == [3, 4]).all()

        assert diff.diff_total == 15
        assert np.isclose(diff.diff_ratio, 0.682, atol=1e-3, rtol=0)

        report = diff.report()
        assert "Column A data differs in row 0:\n    a> True\n    b> False" in report
        assert "...and at 1 more indices.\n Column D data differs in row 0:" in report
        assert "15 different table data element(s) found (68.18% different)" in report
        assert report.count("more indices") == 1

    def test_identical_files_basic(self):
        """Test identicality of two simple, extensionless files."""

        a = np.arange(100).reshape(10, 10)
        hdu = PrimaryHDU(data=a)
        hdu.writeto(self.temp("testa.fits"))
        hdu.writeto(self.temp("testb.fits"))
        diff = FITSDiff(self.temp("testa.fits"), self.temp("testb.fits"))
        assert diff.identical

        report = diff.report()
        # Primary HDUs should contain no differences
        assert "Primary HDU" not in report
        assert "Extension HDU" not in report
        assert "No differences found." in report

        a = np.arange(10)
        ehdu = ImageHDU(data=a)
        diff = HDUDiff(ehdu, ehdu)
        assert diff.identical
        report = diff.report()
        assert "No differences found." in report

    def test_partially_identical_files1(self):
        """
        Test files that have some identical HDUs but a different extension
        count.
        """

        a = np.arange(100).reshape(10, 10)
        phdu = PrimaryHDU(data=a)
        ehdu = ImageHDU(data=a)
        hdula = HDUList([phdu, ehdu])
        hdulb = HDUList([phdu, ehdu, ehdu])
        diff = FITSDiff(hdula, hdulb)
        assert not diff.identical
        assert diff.diff_hdu_count == (2, 3)

        # diff_hdus should be empty, since the third extension in hdulb
        # has nothing to compare against
        assert diff.diff_hdus == []

        report = diff.report()
        assert "Files contain different numbers of HDUs" in report
        assert "a: 2\n b: 3" in report
        assert "No differences found between common HDUs" in report

    def test_partially_identical_files2(self):
        """
        Test files that have some identical HDUs but one different HDU.
        """

        a = np.arange(100).reshape(10, 10)
        phdu = PrimaryHDU(data=a)
        ehdu = ImageHDU(data=a)
        ehdu2 = ImageHDU(data=(a + 1))
        hdula = HDUList([phdu, ehdu, ehdu])
        hdulb = HDUList([phdu, ehdu2, ehdu])
        diff = FITSDiff(hdula, hdulb)

        assert not diff.identical
        assert diff.diff_hdu_count == ()
        assert len(diff.diff_hdus) == 1
        assert diff.diff_hdus[0][0] == 1

        hdudiff = diff.diff_hdus[0][1]
        assert not hdudiff.identical
        assert hdudiff.diff_extnames == ()
        assert hdudiff.diff_extvers == ()
        assert hdudiff.diff_extension_types == ()
        assert hdudiff.diff_headers.identical
        assert hdudiff.diff_data is not None

        datadiff = hdudiff.diff_data
        assert isinstance(datadiff, ImageDataDiff)
        assert not datadiff.identical
        assert datadiff.diff_dimensions == ()
        assert datadiff.diff_pixels == [((0, y), (y, y + 1)) for y in range(10)]
        assert datadiff.diff_ratio == 1.0
        assert datadiff.diff_total == 100

        report = diff.report()
        # Primary HDU and 2nd extension HDU should have no differences
        assert "Primary HDU" not in report
        assert "Extension HDU 2" not in report
        assert "Extension HDU 1" in report

        assert "Headers contain differences" not in report
        assert "Data contains differences" in report
        for y in range(10):
            assert f"Data differs at [{y + 1}, 1]" in report
        assert "100 different pixels found (100.00% different)." in report

    def test_partially_identical_files3(self):
        """
        Test files that have some identical HDUs but a different extension
        name.
        """

        phdu = PrimaryHDU()
        ehdu = ImageHDU(name="FOO")
        hdula = HDUList([phdu, ehdu])
        ehdu = BinTableHDU(name="BAR")
        ehdu.header["EXTVER"] = 2
        ehdu.header["EXTLEVEL"] = 3
        hdulb = HDUList([phdu, ehdu])
        diff = FITSDiff(hdula, hdulb)
        assert not diff.identical

        assert diff.diff_hdus[0][0] == 1

        hdu_diff = diff.diff_hdus[0][1]
        assert hdu_diff.diff_extension_types == ("IMAGE", "BINTABLE")
        assert hdu_diff.diff_extnames == ("FOO", "BAR")
        assert hdu_diff.diff_extvers == (1, 2)
        assert hdu_diff.diff_extlevels == (1, 3)

        report = diff.report()
        assert "Extension types differ" in report
        assert "a: IMAGE\n    b: BINTABLE" in report
        assert "Extension names differ" in report
        assert "a: FOO\n    b: BAR" in report
        assert "Extension versions differ" in report
        assert "a: 1\n    b: 2" in report
        assert "Extension levels differ" in report
        assert "a: 1\n    b: 2" in report

    def test_diff_nans(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/204
        """

        # First test some arrays that should be equivalent....
        arr = np.empty((10, 10), dtype=np.float64)
        arr[:5] = 1.0
        arr[5:] = np.nan
        arr2 = arr.copy()

        table = np.rec.array(
            [(1.0, 2.0), (3.0, np.nan), (np.nan, np.nan)], names=["cola", "colb"]
        ).view(fits.FITS_rec)
        table2 = table.copy()

        assert ImageDataDiff(arr, arr2).identical
        assert TableDataDiff(table, table2).identical

        # Now let's introduce some differences, where there are nans and where
        # there are not nans
        arr2[0][0] = 2.0
        arr2[5][0] = 2.0
        table2[0][0] = 2.0
        table2[1][1] = 2.0

        diff = ImageDataDiff(arr, arr2)
        assert not diff.identical
        assert diff.diff_pixels[0] == ((0, 0), (1.0, 2.0))
        assert diff.diff_pixels[1][0] == (5, 0)
        assert np.isnan(diff.diff_pixels[1][1][0])
        assert diff.diff_pixels[1][1][1] == 2.0

        diff = TableDataDiff(table, table2)
        assert not diff.identical
        assert diff.diff_values[0] == (("cola", 0), (1.0, 2.0))
        assert diff.diff_values[1][0] == ("colb", 1)
        assert np.isnan(diff.diff_values[1][1][0])
        assert diff.diff_values[1][1][1] == 2.0

    def test_file_output_from_path_string(self):
        outpath = self.temp("diff_output.txt")
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        diffobj = HeaderDiff(ha, hb)
        diffobj.report(fileobj=outpath)
        report_as_string = diffobj.report()
        with open(outpath) as fout:
            assert fout.read() == report_as_string

    def test_file_output_overwrite_safety(self):
        outpath = self.temp("diff_output.txt")
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        diffobj = HeaderDiff(ha, hb)
        diffobj.report(fileobj=outpath)

        with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
            diffobj.report(fileobj=outpath)

    def test_file_output_overwrite_success(self):
        outpath = self.temp("diff_output.txt")
        ha = Header([("A", 1), ("B", 2), ("C", 3)])
        hb = ha.copy()
        hb["C"] = 4
        diffobj = HeaderDiff(ha, hb)
        diffobj.report(fileobj=outpath)
        report_as_string = diffobj.report()
        diffobj.report(fileobj=outpath, overwrite=True)
        with open(outpath) as fout:
            assert fout.read() == report_as_string, (
                "overwritten output file is not identical to report string"
            )

    def test_rawdatadiff_nodiff(self):
        a = np.arange(100, dtype="uint8").reshape(10, 10)
        b = a.copy()
        hdu_a = DummyNonstandardExtHDU(data=a)
        hdu_b = DummyNonstandardExtHDU(data=b)
        diff = HDUDiff(hdu_a, hdu_b)
        assert diff.identical
        report = diff.report()
        assert "No differences found." in report

    def test_rawdatadiff_dimsdiff(self):
        a = np.arange(100, dtype="uint8") + 10
        b = a[:80].copy()
        hdu_a = DummyNonstandardExtHDU(data=a)
        hdu_b = DummyNonstandardExtHDU(data=b)
        diff = HDUDiff(hdu_a, hdu_b)
        assert not diff.identical
        report = diff.report()
        assert "Data sizes differ:" in report
        assert "a: 100 bytes" in report
        assert "b: 80 bytes" in report
        assert "No further data comparison performed." in report

    def test_rawdatadiff_bytesdiff(self):
        a = np.arange(100, dtype="uint8") + 10
        b = a.copy()
        changes = [(30, 200), (89, 170)]
        for i, v in changes:
            b[i] = v

        hdu_a = DummyNonstandardExtHDU(data=a)
        hdu_b = DummyNonstandardExtHDU(data=b)
        diff = HDUDiff(hdu_a, hdu_b)

        assert not diff.identical

        diff_bytes = diff.diff_data.diff_bytes
        assert len(changes) == len(diff_bytes)
        for j, (i, v) in enumerate(changes):
            assert diff_bytes[j] == (i, (i + 10, v))

        report = diff.report()
        assert "Data contains differences:" in report
        for i, _ in changes:
            assert f"Data differs at byte {i}:" in report
        assert "2 different bytes found (2.00% different)." in report


def test_fitsdiff_hdu_name(tmp_path):
    """Make sure diff report reports HDU name and ver if same in files"""
    path1 = tmp_path / "test1.fits"
    path2 = tmp_path / "test2.fits"

    hdulist = HDUList([PrimaryHDU(), ImageHDU(data=np.zeros(5), name="SCI")])
    hdulist.writeto(path1)
    hdulist[1].data[0] = 1
    hdulist.writeto(path2)

    diff = FITSDiff(path1, path2)
    assert "Extension HDU 1 (SCI, 1):" in diff.report()


def test_fitsdiff_no_hdu_name(tmp_path):
    """Make sure diff report doesn't report HDU name if not in files"""
    path1 = tmp_path / "test1.fits"
    path2 = tmp_path / "test2.fits"

    hdulist = HDUList([PrimaryHDU(), ImageHDU(data=np.zeros(5))])
    hdulist.writeto(path1)
    hdulist[1].data[0] = 1
    hdulist.writeto(path2)

    diff = FITSDiff(path1, path2)
    assert "Extension HDU 1:" in diff.report()


def test_fitsdiff_with_names(tmp_path):
    """Make sure diff report doesn't report HDU name if not same in files"""
    path1 = tmp_path / "test1.fits"
    path2 = tmp_path / "test2.fits"

    hdulist = HDUList([PrimaryHDU(), ImageHDU(data=np.zeros(5), name="SCI", ver=1)])
    hdulist.writeto(path1)
    hdulist[1].name = "ERR"
    hdulist.writeto(path2)

    diff = FITSDiff(path1, path2)
    assert "Extension HDU 1:" in diff.report()


def test_rawdatadiff_diff_with_rtol(tmp_path):
    """Regression test for https://github.com/astropy/astropy/issues/13330"""
    path1 = tmp_path / "test1.fits"
    path2 = tmp_path / "test2.fits"
    a = np.zeros((10, 2), dtype="float32")
    a[:, 0] = np.arange(10, dtype="float32") + 10
    a[:, 1] = np.arange(10, dtype="float32") + 20
    b = a.copy()
    changes = [(3, 13.1, 23.1), (8, 20.5, 30.5)]
    for i, v, w in changes:
        b[i, 0] = v
        b[i, 1] = w

    ca = Column("A", format="20E", array=[a])
    cb = Column("A", format="20E", array=[b])
    hdu_a = BinTableHDU.from_columns([ca])
    hdu_a.writeto(path1, overwrite=True)
    hdu_b = BinTableHDU.from_columns([cb])
    hdu_b.writeto(path2, overwrite=True)
    with fits.open(path1) as fits1:
        with fits.open(path2) as fits2:
            diff = FITSDiff(fits1, fits2, atol=0, rtol=0.001)
            str1 = diff.report(fileobj=None, indent=0)

            diff = FITSDiff(fits1, fits2, atol=0, rtol=0.01)
            str2 = diff.report(fileobj=None, indent=0)

    assert "...and at 1 more indices." in str1
    assert "...and at 1 more indices." not in str2



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections import defaultdict

import numpy as np
from matplotlib import rcParams
from matplotlib.lines import Line2D, Path
from matplotlib.transforms import Affine2D


class Ticks(Line2D):
    """
    Ticks are derived from Line2D, and note that ticks themselves
    are markers. Thus, you should use set_mec, set_mew, etc.

    To change the tick size (length), you need to use
    set_ticksize. To change the direction of the ticks (ticks are
    in opposite direction of ticklabels by default), use
    set_tick_out(False).

    Note that Matplotlib's defaults dictionary :data:`~matplotlib.rcParams`
    contains default settings (color, size, width) of the form `xtick.*` and
    `ytick.*`. In a WCS projection, there may not be a clear relationship
    between axes of the projection and 'x' or 'y' axes. For this reason,
    we read defaults from `xtick.*`. The following settings affect the
    default appearance of ticks:

    * `xtick.direction`
    * `xtick.major.size`
    * `xtick.major.width`
    * `xtick.minor.size`
    * `xtick.color`

    Attributes
    ----------
    ticks_locs : dict
        This is set when the ticks are drawn, and is a mapping from axis to
        the locations of the ticks for that axis.
    """

    def __init__(self, frame=None, ticksize=None, **kwargs):
        self._frame = frame
        if ticksize is None:
            ticksize = rcParams["xtick.major.size"]
        self.set_ticksize(ticksize)
        self.set_minor_ticksize(rcParams["xtick.minor.size"])
        self.set_tick_out(rcParams["xtick.direction"] == "out")
        self.clear()
        line2d_kwargs = {
            "color": rcParams["xtick.color"],
            "linewidth": rcParams["xtick.major.width"],
        }
        line2d_kwargs.update(kwargs)
        Line2D.__init__(self, [0.0], [0.0], **line2d_kwargs)
        self.set_visible_axes("all")
        self._display_minor_ticks = False

    def display_minor_ticks(self, display_minor_ticks):
        self._display_minor_ticks = display_minor_ticks

    def get_display_minor_ticks(self):
        return self._display_minor_ticks

    def set_tick_out(self, tick_out):
        """
        set True if tick need to be rotated by 180 degree.
        """
        self._tick_out = tick_out

    def get_tick_out(self):
        """
        Return True if the tick will be rotated by 180 degree.
        """
        return self._tick_out

    def set_ticksize(self, ticksize):
        """
        set length of the ticks in points.
        """
        self._ticksize = ticksize

    def get_ticksize(self):
        """
        Return length of the ticks in points.
        """
        return self._ticksize

    def set_minor_ticksize(self, ticksize):
        """
        set length of the minor ticks in points.
        """
        self._minor_ticksize = ticksize

    def get_minor_ticksize(self):
        """
        Return length of the minor ticks in points.
        """
        return self._minor_ticksize

    @property
    def out_size(self):
        if self._tick_out:
            return self._ticksize
        else:
            return 0.0

    def set_visible_axes(self, visible_axes):
        self._visible_axes = self._frame._validate_positions(visible_axes)

    def get_visible_axes(self):
        if self._visible_axes == "all":
            return list(self._frame.keys())
        else:
            return [x for x in self._visible_axes if x in self._frame or x == "#"]

    def clear(self):
        self.world = defaultdict(list)
        self.pixel = defaultdict(list)
        self.angle = defaultdict(list)
        self.disp = defaultdict(list)
        self.minor_world = defaultdict(list)
        self.minor_pixel = defaultdict(list)
        self.minor_angle = defaultdict(list)
        self.minor_disp = defaultdict(list)

    def add(self, axis, world, pixel, angle, axis_displacement):
        self.world[axis].append(world)
        self.pixel[axis].append(pixel)
        self.angle[axis].append(angle)
        self.disp[axis].append(axis_displacement)

    def get_minor_world(self):
        return self.minor_world

    def add_minor(
        self, minor_axis, minor_world, minor_pixel, minor_angle, minor_axis_displacement
    ):
        self.minor_world[minor_axis].append(minor_world)
        self.minor_pixel[minor_axis].append(minor_pixel)
        self.minor_angle[minor_axis].append(minor_angle)
        self.minor_disp[minor_axis].append(minor_axis_displacement)

    def __len__(self):
        return len(self.world)

    _tickvert_path = Path([[0.0, 0.0], [1.0, 0.0]])

    def draw(self, renderer):
        """
        Draw the ticks.
        """
        self.ticks_locs = defaultdict(list)

        if not self.get_visible():
            return

        offset = renderer.points_to_pixels(self.get_ticksize())
        self._draw_ticks(renderer, self.pixel, self.angle, offset)
        if self._display_minor_ticks:
            offset = renderer.points_to_pixels(self.get_minor_ticksize())
            self._draw_ticks(renderer, self.minor_pixel, self.minor_angle, offset)

    def _draw_ticks(self, renderer, pixel_array, angle_array, offset):
        """
        Draw the minor ticks.
        """
        path_trans = self.get_transform()

        gc = renderer.new_gc()
        gc.set_foreground(self.get_color())
        gc.set_alpha(self.get_alpha())
        gc.set_linewidth(self.get_linewidth())

        marker_scale = Affine2D().scale(offset, offset)
        marker_rotation = Affine2D()
        marker_transform = marker_scale + marker_rotation

        initial_angle = 180.0 if self.get_tick_out() else 0.0

        for axis in self.get_visible_axes():
            if axis == "#":
                continue

            if axis not in pixel_array:
                continue

            for loc, angle in zip(pixel_array[axis], angle_array[axis]):
                # Set the rotation for this tick
                marker_rotation.rotate_deg(initial_angle + angle)

                # Draw the markers
                locs = path_trans.transform_non_affine(np.array([loc, loc]))
                renderer.draw_markers(
                    gc,
                    self._tickvert_path,
                    marker_transform,
                    Path(locs),
                    path_trans.get_affine(),
                )

                # Reset the tick rotation before moving to the next tick
                marker_rotation.clear()

                self.ticks_locs[axis].append(locs)

        gc.restore()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from functools import partial

import numpy as np

from astropy import units as u
from astropy.modeling.convolution import Convolution
from astropy.modeling.core import SPECIAL_OPERATORS, CompoundModel
from astropy.nddata import support_nddata
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.console import human_file_size
from astropy.utils.exceptions import AstropyUserWarning

from ._convolve import _convolveNd_c
from .core import MAX_NORMALIZATION, Kernel, Kernel1D, Kernel2D
from .utils import KernelSizeError, has_even_axis

# np.unique([scipy.fft.next_fast_len(i, real=True) for i in range(10000)])
# fmt: off
_good_sizes = np.array(
    [
        0,    1,    2,    3,    4,    5,    6,    8,    9,    10,   12,
        15,   16,   18,   20,   24,   25,   27,   30,   32,   36,   40,
        45,   48,   50,   54,   60,   64,   72,   75,   80,   81,   90,
        96,   100,  108,  120,  125,  128,  135,  144,  150,  160,  162,
        180,  192,  200,  216,  225,  240,  243,  250,  256,  270,  288,
        300,  320,  324,  360,  375,  384,  400,  405,  432,  450,  480,
        486,  500,  512,  540,  576,  600,  625,  640,  648,  675,  720,
        729,  750,  768,  800,  810,  864,  900,  960,  972,  1000, 1024,
        1080, 1125, 1152, 1200, 1215, 1250, 1280, 1296, 1350, 1440, 1458,
        1500, 1536, 1600, 1620, 1728, 1800, 1875, 1920, 1944, 2000, 2025,
        2048, 2160, 2187, 2250, 2304, 2400, 2430, 2500, 2560, 2592, 2700,
        2880, 2916, 3000, 3072, 3125, 3200, 3240, 3375, 3456, 3600, 3645,
        3750, 3840, 3888, 4000, 4050, 4096, 4320, 4374, 4500, 4608, 4800,
        4860, 5000, 5120, 5184, 5400, 5625, 5760, 5832, 6000, 6075, 6144,
        6250, 6400, 6480, 6561, 6750, 6912, 7200, 7290, 7500, 7680, 7776,
        8000, 8100, 8192, 8640, 8748, 9000, 9216, 9375, 9600, 9720, 10000,
    ]
)
# fmt: on
_good_range = int(np.log10(_good_sizes[-1]))

# Disabling doctests when scipy isn't present.
__doctest_requires__ = {("convolve_fft",): ["scipy.fft"]}

BOUNDARY_OPTIONS = [None, "fill", "wrap", "extend"]


def _next_fast_lengths(shape):
    """
    Find optimal or good sizes to pad an array of ``shape`` to for better
    performance with `numpy.fft.*fft` and `scipy.fft.*fft`.
    Calculated directly with `scipy.fft.next_fast_len`, if available; otherwise
    looked up from list and scaled by powers of 10, if necessary.
    """
    if HAS_SCIPY:
        import scipy.fft

        return np.array([scipy.fft.next_fast_len(j) for j in shape])

    newshape = np.empty(len(np.atleast_1d(shape)), dtype=int)
    for i, j in enumerate(shape):
        scale = 10 ** max(int(np.ceil(np.log10(j))) - _good_range, 0)
        for n in _good_sizes:
            if n * scale >= j:
                newshape[i] = n * scale
                break
        else:
            raise ValueError(
                f"No next fast length for {j} found in list of _good_sizes "
                f"<= {_good_sizes[-1] * scale}."
            )
    return newshape


def _copy_input_if_needed(
    input, dtype=float, order="C", nan_treatment=None, mask=None, fill_value=None
):
    # Alias input
    input = input.array if isinstance(input, Kernel) else input
    # strip quantity attributes
    if hasattr(input, "unit"):
        input = input.value
    output = input
    # Copy input
    try:
        # Anything that's masked must be turned into NaNs for the interpolation.
        # This requires copying. A copy is also needed for nan_treatment == 'fill'
        # A copy prevents possible function side-effects of the input array.
        if nan_treatment == "fill" or np.ma.is_masked(input) or mask is not None:
            if np.ma.is_masked(input):
                # ``np.ma.maskedarray.filled()`` returns a copy, however there
                # is no way to specify the return type or order etc. In addition
                # ``np.nan`` is a ``float`` and there is no conversion to an
                # ``int`` type. Therefore, a pre-fill copy is needed for non
                # ``float`` masked arrays. ``asanyarray`` is needed to retain
                # ``np.ma.maskedarray.filled()``.
                # A copy is made if and only if order isn't already correct.
                output = np.asanyarray(input, dtype=dtype, order=order)
                output = output.filled(fill_value)
            else:
                # Since we're making a copy, we might as well use `subok=False` to save,
                # what is probably, a negligible amount of memory.
                output = np.array(
                    input, dtype=dtype, copy=True, order=order, subok=False
                )

            if mask is not None:
                # mask != 0 yields a bool mask for all ints/floats/bool
                output[mask != 0] = fill_value
        else:
            output = np.asanyarray(input, dtype=dtype, order=order)
    except (TypeError, ValueError) as e:
        raise TypeError(
            "input should be a Numpy array or something convertible into a float array",
            e,
        )
    return output


@support_nddata(data="array")
def convolve(
    array,
    kernel,
    boundary="fill",
    fill_value=0.0,
    nan_treatment="interpolate",
    normalize_kernel=True,
    mask=None,
    preserve_nan=False,
    normalization_zero_tol=1e-8,
):
    """
    Convolve an array with a kernel.

    This routine differs from `scipy.ndimage.convolve` because
    it includes a special treatment for ``NaN`` values. Rather
    than including ``NaN`` values in the array in the convolution
    calculation, which causes large ``NaN`` holes in the convolved
    array, ``NaN`` values are replaced with interpolated values using
    the kernel as an interpolation function.

    Parameters
    ----------
    array : `~astropy.nddata.NDData` or array-like
        The array to convolve. This should be a 1, 2, or 3-dimensional
        array or a list or a set of nested lists representing a 1, 2,
        or 3-dimensional array. If an `~astropy.nddata.NDData`, the
        ``mask`` of the `~astropy.nddata.NDData` will be used as the
        ``mask`` argument.
    kernel : `numpy.ndarray` or `~astropy.convolution.Kernel`
        The convolution kernel. The number of dimensions should match
        those for the array, and the dimensions should be odd in all
        directions. If a masked array, the masked values will be
        replaced by ``fill_value``.
    boundary : str, optional
        A flag indicating how to handle boundaries:

        * `None`: Set the ``result`` values to zero where the kernel
          extends beyond the edge of the array.
        * 'fill': Set values outside the array boundary to
          ``fill_value`` (default).
        * 'wrap': Periodic boundary that wrap to the other side of
          ``array``.
        * 'extend': Set values outside the array to the nearest
          ``array`` value.
    fill_value : float, optional
        The value to use outside the array when using
        ``boundary='fill'``.
    normalize_kernel : bool, optional
        Whether to normalize the kernel to have a sum of one.
    nan_treatment : {'interpolate', 'fill'}, optional
        The method used to handle NaNs in the input ``array``:

        * 'interpolate': ``NaN`` values are replaced with interpolated
          values using the kernel as an interpolation function. Note that
          if the kernel has a sum equal to zero, NaN interpolation is not
          possible and will raise an exception.
        * 'fill': ``NaN`` values are replaced by ``fill_value`` prior to
          convolution.
    preserve_nan : bool, optional
        After performing convolution, should pixels that were originally
        NaN again become NaN?
    mask : None or ndarray, optional
        A "mask" array. Shape must match ``array``, and anything that
        is masked (i.e., not 0/`False`) will be set to NaN for the
        convolution. If `None`, no masking will be performed unless
        ``array`` is a masked array. If ``mask`` is not `None` *and*
        ``array`` is a masked array, a pixel is masked if it is masked
        in either ``mask`` *or* ``array.mask``.
    normalization_zero_tol : float, optional
        The absolute tolerance on whether the kernel is different than
        zero. If the kernel sums to zero to within this precision, it
        cannot be normalized. Default is "1e-8".

    Returns
    -------
    result : `numpy.ndarray`
        An array with the same dimensions and as the input array,
        convolved with kernel. The data type depends on the input array
        type. If array is a floating point type, then the return array
        keeps the same data type, otherwise the type is ``numpy.float``.

    Notes
    -----
    For masked arrays, masked values are treated as NaNs. The
    convolution is always done at ``numpy.float`` precision.
    """
    if boundary not in BOUNDARY_OPTIONS:
        raise ValueError(f"Invalid boundary option: must be one of {BOUNDARY_OPTIONS}")

    if nan_treatment not in ("interpolate", "fill"):
        raise ValueError("nan_treatment must be one of 'interpolate','fill'")

    # OpenMP support is disabled at the C src code level, changing this will have
    # no effect.
    n_threads = 1

    # Keep refs to originals
    passed_kernel = kernel
    passed_array = array

    # The C routines all need float type inputs (so, a particular
    # bit size, endianness, etc.).  So we have to convert, which also
    # has the effect of making copies so we don't modify the inputs.
    # After this, the variables we work with will be array_internal, and
    # kernel_internal.  However -- we do want to keep track of what type
    # the input array was so we can cast the result to that at the end
    # if it's a floating point type.  Don't bother with this for lists --
    # just always push those as float.
    # It is always necessary to make a copy of kernel (since it is modified),
    # but, if we just so happen to be lucky enough to have the input array
    # have exactly the desired type, we just alias to array_internal
    # Convert kernel to ndarray if not already

    # Copy or alias array to array_internal
    array_internal = _copy_input_if_needed(
        passed_array,
        dtype=float,
        order="C",
        nan_treatment=nan_treatment,
        mask=mask,
        fill_value=np.nan,
    )
    array_dtype = getattr(passed_array, "dtype", array_internal.dtype)
    # Copy or alias kernel to kernel_internal
    kernel_internal = _copy_input_if_needed(
        passed_kernel,
        dtype=float,
        order="C",
        nan_treatment=None,
        mask=None,
        fill_value=fill_value,
    )

    # Make sure kernel has all odd axes
    if has_even_axis(kernel_internal):
        raise KernelSizeError("Kernel size must be odd in all axes.")

    # If both image array and kernel are Kernel instances
    # constrain convolution method
    # This must occur before the main alias/copy of ``passed_kernel`` to
    # ``kernel_internal`` as it is used for filling masked kernels.
    if isinstance(passed_array, Kernel) and isinstance(passed_kernel, Kernel):
        warnings.warn(
            "Both array and kernel are Kernel instances, hardwiring "
            "the following parameters: boundary='fill', fill_value=0,"
            " normalize_Kernel=True, nan_treatment='interpolate'",
            AstropyUserWarning,
        )
        boundary = "fill"
        fill_value = 0
        normalize_kernel = True
        nan_treatment = "interpolate"

    # -----------------------------------------------------------------------
    # From this point onwards refer only to ``array_internal`` and
    # ``kernel_internal``.
    # Assume both are base np.ndarrays and NOT subclasses e.g. NOT
    # ``Kernel`` nor ``np.ma.maskedarray`` classes.
    # -----------------------------------------------------------------------

    # Check dimensionality
    if array_internal.ndim == 0:
        raise ValueError("cannot convolve 0-dimensional arrays")
    elif array_internal.ndim > 3:
        raise NotImplementedError(
            "convolve only supports 1, 2, and 3-dimensional arrays at this time"
        )
    elif array_internal.ndim != kernel_internal.ndim:
        raise ValueError("array and kernel have differing number of dimensions.")
    elif array_internal.size == 0:
        raise ValueError("cannot convolve empty array")

    array_shape = np.array(array_internal.shape)
    kernel_shape = np.array(kernel_internal.shape)
    pad_width = kernel_shape // 2

    # For boundary=None only the center space is convolved. All array indices within a
    # distance kernel.shape//2 from the edge are completely ignored (zeroed).
    # E.g. (1D list) only the indices len(kernel)//2 : len(array)-len(kernel)//2
    # are convolved. It is therefore not possible to use this method to convolve an
    # array by a kernel that is larger (see note below) than the array - as ALL pixels
    # would be ignored leaving an array of only zeros.
    # Note: For even kernels the correctness condition is array_shape > kernel_shape.
    # For odd kernels it is:
    #     array_shape >= kernel_shape OR
    #     array_shape > kernel_shape-1 OR
    #     array_shape > 2*(kernel_shape//2).
    # Since the latter is equal to the former two for even lengths, the latter condition is
    # complete.
    if boundary is None and not np.all(array_shape > 2 * pad_width):
        raise KernelSizeError(
            "for boundary=None all kernel axes must be smaller than array's - "
            "use boundary in ['fill', 'extend', 'wrap'] instead."
        )

    # NaN interpolation significantly slows down the C convolution
    # computation. Since nan_treatment = 'interpolate', is the default
    # check whether it is even needed, if not, don't interpolate.
    # NB: np.isnan(array_internal.sum()) is faster than np.isnan(array_internal).any()
    nan_interpolate = (nan_treatment == "interpolate") and np.isnan(
        array_internal.sum()
    )

    # Check if kernel is normalizable
    if normalize_kernel or nan_interpolate:
        kernel_sum = kernel_internal.sum()
        kernel_sums_to_zero = np.isclose(kernel_sum, 0, atol=normalization_zero_tol)

        if kernel_sum < 1.0 / MAX_NORMALIZATION or kernel_sums_to_zero:
            if nan_interpolate:
                raise ValueError(
                    "Setting nan_treatment='interpolate' "
                    "requires the kernel to be normalized, "
                    "but the input kernel has a sum close "
                    "to zero. For a zero-sum kernel and "
                    "data with NaNs, set nan_treatment='fill'."
                )
            else:
                raise ValueError(
                    "The kernel can't be normalized, because "
                    "its sum is close to zero. The sum of the "
                    f"given kernel is < {1.0 / MAX_NORMALIZATION:.2f}. "
                    "For a zero-sum kernel, set normalize_kernel=False "
                    "or pass a custom normalization function to "
                    "normalize_kernel."
                )

    # Mark the NaN values so we can replace them later if interpolate_nan is
    # not set
    if preserve_nan or nan_treatment == "fill":
        initially_nan = np.isnan(array_internal)
        if nan_treatment == "fill":
            array_internal[initially_nan] = fill_value

    # Avoid any memory allocation within the C code. Allocate output array
    # here and pass through instead.
    result = np.zeros(array_internal.shape, dtype=float, order="C")

    embed_result_within_padded_region = True
    array_to_convolve = array_internal
    if boundary in ("fill", "extend", "wrap"):
        embed_result_within_padded_region = False
        if boundary == "fill":
            # This method is faster than using numpy.pad(..., mode='constant')
            array_to_convolve = np.full(
                array_shape + 2 * pad_width,
                fill_value=fill_value,
                dtype=float,
                order="C",
            )
            # Use bounds [pad_width[0]:array_shape[0]+pad_width[0]] instead of
            #            [pad_width[0]:-pad_width[0]]
            # to account for when the kernel has size of 1 making pad_width = 0.
            if array_internal.ndim == 1:
                array_to_convolve[pad_width[0] : array_shape[0] + pad_width[0]] = (
                    array_internal
                )
            elif array_internal.ndim == 2:
                array_to_convolve[
                    pad_width[0] : array_shape[0] + pad_width[0],
                    pad_width[1] : array_shape[1] + pad_width[1],
                ] = array_internal
            else:
                array_to_convolve[
                    pad_width[0] : array_shape[0] + pad_width[0],
                    pad_width[1] : array_shape[1] + pad_width[1],
                    pad_width[2] : array_shape[2] + pad_width[2],
                ] = array_internal
        else:
            np_pad_mode_dict = {"fill": "constant", "extend": "edge", "wrap": "wrap"}
            np_pad_mode = np_pad_mode_dict[boundary]
            pad_width = kernel_shape // 2

            if array_internal.ndim == 1:
                np_pad_width = (pad_width[0],)
            elif array_internal.ndim == 2:
                np_pad_width = ((pad_width[0],), (pad_width[1],))
            else:
                np_pad_width = ((pad_width[0],), (pad_width[1],), (pad_width[2],))

            array_to_convolve = np.pad(
                array_internal, pad_width=np_pad_width, mode=np_pad_mode
            )

    _convolveNd_c(
        result,
        array_to_convolve,
        kernel_internal,
        nan_interpolate,
        embed_result_within_padded_region,
        n_threads,
    )

    # So far, normalization has only occurred for nan_treatment == 'interpolate'
    # because this had to happen within the C extension so as to ignore
    # any NaNs
    if normalize_kernel:
        if not nan_interpolate:
            result /= kernel_sum
    elif nan_interpolate:
        result *= kernel_sum

    if nan_interpolate and not preserve_nan and np.isnan(result.sum()):
        warnings.warn(
            "nan_treatment='interpolate', however, NaN values detected "
            "post convolution. A contiguous region of NaN values, larger "
            "than the kernel size, are present in the input array. "
            "Increase the kernel size to avoid this.",
            AstropyUserWarning,
        )

    if preserve_nan:
        result[initially_nan] = np.nan

    # Convert result to original data type
    array_unit = getattr(passed_array, "unit", None)
    if array_unit is not None:
        result <<= array_unit

    if isinstance(passed_array, Kernel):
        if isinstance(passed_array, Kernel1D):
            new_result = Kernel1D(array=result)
        elif isinstance(passed_array, Kernel2D):
            new_result = Kernel2D(array=result)
        else:
            raise TypeError("Only 1D and 2D Kernels are supported.")
        new_result._is_bool = False
        new_result._separable = passed_array._separable
        if isinstance(passed_kernel, Kernel):
            new_result._separable = new_result._separable and passed_kernel._separable
        return new_result
    if array_dtype.kind == "f":
        # Try to preserve the input type if it's a floating point type
        return result.astype(array_dtype, copy=False)
    else:
        return result


@support_nddata(data="array")
def convolve_fft(
    array,
    kernel,
    boundary="fill",
    fill_value=0.0,
    nan_treatment="interpolate",
    normalize_kernel=True,
    normalization_zero_tol=1e-8,
    preserve_nan=False,
    mask=None,
    crop=True,
    return_fft=False,
    fft_pad=None,
    psf_pad=None,
    min_wt=0.0,
    allow_huge=False,
    fftn=np.fft.fftn,
    ifftn=np.fft.ifftn,
    complex_dtype=complex,
    dealias=False,
):
    """
    Convolve an ndarray with an nd-kernel.

    Returns a convolved image with ``shape = array.shape``. Assumes
    kernel is centered.

    `convolve_fft` is very similar to `convolve` in that it replaces
    ``NaN`` values in the original image with interpolated values using
    the kernel as an interpolation function. However, it also includes
    many additional options specific to the implementation.

    `convolve_fft` differs from `scipy.signal.fftconvolve` in a few
    ways:

    * It can treat ``NaN`` values as zeros or interpolate over them.
    * ``inf`` values are treated as ``NaN``
    * It optionally pads to the nearest faster sizes to improve
      FFT speed. These sizes are optimized for the numpy and scipy
      implementations, and `~scipy.signal.fftconvolve` uses them by
      default as well; when using other external functions (see below),
      results may vary.
    * It's only valid ``mode`` is 'same' (i.e., the same shape array is
      returned)
    * It lets you use your own fft, e.g., `pyFFTW
      <https://pypi.org/project/pyFFTW/>`_ or `pyFFTW3
      <https://pypi.org/project/PyFFTW3/0.2.1/>`_, which can lead to
      performance improvements, depending on your system configuration.
      pyFFTW3 is threaded, and therefore may yield significant
      performance benefits on multicore machines at the cost of
      greater memory requirements. Specify the ``fftn`` and ``ifftn``
      keywords to override the default, which is `numpy.fft.fftn` and
      `numpy.fft.ifftn`. The `scipy.fft` functions also offer somewhat
      better performance and a multithreaded option.

    Parameters
    ----------
    array : `numpy.ndarray`
        Array to be convolved with ``kernel``. It can be of any
        dimensionality, though only 1, 2, and 3d arrays have been
        tested.
    kernel : `numpy.ndarray` or `astropy.convolution.Kernel`
        The convolution kernel. The number of dimensions should match
        those for the array. The dimensions *do not* have to be odd in
        all directions, unlike in the non-fft `convolve` function. The
        kernel will be normalized if ``normalize_kernel`` is set. It is
        assumed to be centered (i.e., shifts may result if your kernel
        is asymmetric)
    boundary : {'fill', 'wrap'}, optional
        A flag indicating how to handle boundaries:

        * 'fill': set values outside the array boundary to
          ``fill_value`` (default)
        * 'wrap': periodic boundary

        The `None` and 'extend' parameters are not supported for
        FFT-based convolution.
    fill_value : float, optional
        The value to use outside the array when using boundary='fill'.
    nan_treatment : {'interpolate', 'fill'}, optional
        The method used to handle NaNs in the input ``array``:

        * 'interpolate': ``NaN`` values are replaced with interpolated
          values using the kernel as an interpolation function. Note that
          if the kernel has a sum equal to zero, NaN interpolation is not
          possible and will raise an exception.
        * 'fill': ``NaN`` values are replaced by ``fill_value`` prior to
          convolution.
    normalize_kernel : callable or boolean, optional
        If specified, this is the function to divide kernel by
        to normalize it. e.g., ``normalize_kernel=np.sum`` means
        that kernel will be modified to be: ``kernel = kernel /
        np.sum(kernel)``. If True, defaults to ``normalize_kernel =
        np.sum``.
    normalization_zero_tol : float, optional
        The absolute tolerance on whether the kernel is different than
        zero. If the kernel sums to zero to within this precision, it
        cannot be normalized. Default is "1e-8".
    preserve_nan : bool, optional
        After performing convolution, should pixels that were originally
        NaN again become NaN?
    mask : None or ndarray, optional
        A "mask" array. Shape must match ``array``, and anything that
        is masked (i.e., not 0/`False`) will be set to NaN for the
        convolution. If `None`, no masking will be performed unless
        ``array`` is a masked array. If ``mask`` is not `None` *and*
        ``array`` is a masked array, a pixel is masked if it is masked
        in either ``mask`` *or* ``array.mask``.
    crop : bool, optional
        Default on. Return an image of the size of the larger of
        the input image and the kernel. If the image and kernel are
        asymmetric in opposite directions, will return the largest image
        in both directions. For example, if an input image has shape
        [100,3] but a kernel with shape [6,6] is used, the output will
        be [100,6].
    return_fft : bool, optional
        Return the ``fft(image)*fft(kernel)`` instead of the convolution
        (which is ``ifft(fft(image)*fft(kernel))``). Useful for making
        PSDs.
    fft_pad : bool, optional
        Default on. Zero-pad image to the nearest size supporting more
        efficient execution of the FFT, generally values factorizable
        into the first 3-5 prime numbers. With ``boundary='wrap'``, this
        will be disabled.
    psf_pad : bool, optional
        Zero-pad image to be at least the sum of the image sizes to
        avoid edge-wrapping when smoothing. This is enabled by default
        with ``boundary='fill'``, but it can be overridden with a
        boolean option. ``boundary='wrap'`` and ``psf_pad=True`` are not
        compatible.
    min_wt : float, optional
        If ignoring ``NaN`` / zeros, force all grid points with a weight
        less than this value to ``NaN`` (the weight of a grid point with
        *no* ignored neighbors is 1.0). If ``min_wt`` is zero, then all
        zero-weight points will be set to zero instead of ``NaN`` (which
        they would be otherwise, because 1/0 = nan). See the examples
        below.
    allow_huge : bool, optional
        Allow huge arrays in the FFT? If False, will raise an exception
        if the array or kernel size is >1 GB.
    fftn : callable, optional
        The fft function. Can be overridden to use your own ffts, e.g.
        an fftw3 wrapper or scipy's fftn, ``fft=scipy.fftpack.fftn``.
    ifftn : callable, optional
        The inverse fft function. Can be overridden the same way
        ``fttn``.
    complex_dtype : complex type, optional
        Which complex dtype to use. `numpy` has a range of options, from
        64 to 256.
    dealias: bool, optional
        Default off. Zero-pad image to enable explicit dealiasing of
        convolution. With ``boundary='wrap'``, this will be disabled.
        Note that for an input of nd dimensions this will increase the
        size of the temporary arrays by at least ``1.5**nd``. This may
        result in significantly more memory usage.

    Returns
    -------
    default : ndarray
        ``array`` convolved with ``kernel``. If ``return_fft`` is set,
        returns ``fft(array) * fft(kernel)``. If crop is not set,
        returns the image, but with the fft-padded size instead of the
        input size.

    Raises
    ------
    `ValueError`
        If the array is bigger than 1 GB after padding, will raise this
        exception unless ``allow_huge`` is True.

    See Also
    --------
    convolve:
        Convolve is a non-fft version of this code. It is more memory
        efficient and for small kernels can be faster.

    Notes
    -----
    With ``psf_pad=True`` and a large PSF, the resulting data
    can become large and consume a lot of memory. See Issue
    https://github.com/astropy/astropy/pull/4366 and the update in
    https://github.com/astropy/astropy/pull/11533 for further details.

    Dealiasing of pseudospectral convolutions is necessary for
    numerical stability of the underlying algorithms. A common
    method for handling this is to zero pad the image by at least
    1/2 to eliminate the wavenumbers which have been aliased
    by convolution. This is so that the aliased 1/3 of the
    results of the convolution computation can be thrown out. See
    https://doi.org/10.1175/1520-0469(1971)028%3C1074:OTEOAI%3E2.0.CO;2
    https://iopscience.iop.org/article/10.1088/1742-6596/318/7/072037

    Note that if dealiasing is necessary to your application, but your
    process is memory constrained, you may want to consider using
    FFTW++: https://github.com/dealias/fftwpp. It includes python
    wrappers for a pseudospectral convolution which will implicitly
    dealias your convolution without the need for additional padding.
    Note that one cannot use FFTW++'s convlution directly in this
    method as in handles the entire convolution process internally.
    Additionally, FFTW++ includes other useful pseudospectral methods to
    consider.

    Examples
    --------
    >>> convolve_fft([1, 0, 3], [1, 1, 1])
    array([0.33333333, 1.33333333, 1.        ])

    >>> convolve_fft([1, np.nan, 3], [1, 1, 1])
    array([0.5, 2. , 1.5])

    >>> convolve_fft([1, 0, 3], [0, 1, 0])  # doctest: +FLOAT_CMP
    array([ 1.00000000e+00, -3.70074342e-17,  3.00000000e+00])

    >>> convolve_fft([1, 2, 3], [1])
    array([1., 2., 3.])

    >>> convolve_fft([1, np.nan, 3], [0, 1, 0], nan_treatment='interpolate')
    array([1., 0., 3.])

    >>> convolve_fft([1, np.nan, 3], [0, 1, 0], nan_treatment='interpolate',
    ...              min_wt=1e-8)
    array([ 1., nan,  3.])

    >>> convolve_fft([1, np.nan, 3], [1, 1, 1], nan_treatment='interpolate')
    array([0.5, 2. , 1.5])

    >>> convolve_fft([1, np.nan, 3], [1, 1, 1], nan_treatment='interpolate',
    ...               normalize_kernel=True)
    array([0.5, 2. , 1.5])

    >>> import scipy.fft  # optional - requires scipy
    >>> convolve_fft([1, np.nan, 3], [1, 1, 1], nan_treatment='interpolate',
    ...               normalize_kernel=True,
    ...               fftn=scipy.fft.fftn, ifftn=scipy.fft.ifftn)
    array([0.5, 2. , 1.5])

    >>> fft_mp = lambda a: scipy.fft.fftn(a, workers=-1)  # use all available cores
    >>> ifft_mp = lambda a: scipy.fft.ifftn(a, workers=-1)
    >>> convolve_fft([1, np.nan, 3], [1, 1, 1], nan_treatment='interpolate',
    ...               normalize_kernel=True, fftn=fft_mp, ifftn=ifft_mp)
    array([0.5, 2. , 1.5])
    """
    # Checking copied from convolve.py - however, since FFTs have real &
    # complex components, we change the types.  Only the real part will be
    # returned! Note that this always makes a copy.

    # Check kernel is kernel instance
    if isinstance(kernel, Kernel):
        kernel = kernel.array
        if isinstance(array, Kernel):
            raise TypeError(
                "Can't convolve two kernels with convolve_fft.  Use convolve instead."
            )

    if nan_treatment not in ("interpolate", "fill"):
        raise ValueError("nan_treatment must be one of 'interpolate','fill'")

    # Get array quantity if it exists
    array_unit = getattr(array, "unit", None)

    # Convert array dtype to complex
    # and ensure that list inputs become arrays
    array = _copy_input_if_needed(
        array,
        dtype=complex,
        order="C",
        nan_treatment=nan_treatment,
        mask=mask,
        fill_value=np.nan,
    )
    kernel = _copy_input_if_needed(
        kernel, dtype=complex, order="C", nan_treatment=None, mask=None, fill_value=0
    )

    # Check that the number of dimensions is compatible
    if array.ndim != kernel.ndim:
        raise ValueError("Image and kernel must have same number of dimensions")

    arrayshape = array.shape
    kernshape = kernel.shape

    array_size_B = (
        np.prod(arrayshape, dtype=np.int64) * np.dtype(complex_dtype).itemsize
    ) * u.byte
    if array_size_B > 1 * u.GB and not allow_huge:
        raise ValueError(
            f"Size Error: Arrays will be {human_file_size(array_size_B)}.  "
            "Use allow_huge=True to override this exception."
        )

    # NaN and inf catching
    nanmaskarray = np.isnan(array) | np.isinf(array)
    if nan_treatment == "fill":
        array[nanmaskarray] = fill_value
    else:
        array[nanmaskarray] = 0
    nanmaskkernel = np.isnan(kernel) | np.isinf(kernel)
    kernel[nanmaskkernel] = 0

    if normalize_kernel is True:
        if kernel.sum() < 1.0 / MAX_NORMALIZATION:
            raise Exception(
                "The kernel can't be normalized, because its sum is close "
                "to zero. The sum of the given kernel is < "
                f"{1.0 / MAX_NORMALIZATION:.2f}. For a zero-sum kernel, set "
                "normalize_kernel=False or pass a custom normalization "
                "function to normalize_kernel."
            )
        kernel_scale = kernel.sum()
        normalized_kernel = kernel / kernel_scale
        kernel_scale = 1  # if we want to normalize it, leave it normed!
    elif normalize_kernel:
        # try this.  If a function is not passed, the code will just crash... I
        # think type checking would be better but PEPs say otherwise...
        kernel_scale = normalize_kernel(kernel)
        normalized_kernel = kernel / kernel_scale
    else:
        kernel_scale = kernel.sum()
        if np.abs(kernel_scale) < normalization_zero_tol:
            if nan_treatment == "interpolate":
                raise ValueError(
                    "Cannot interpolate NaNs with an unnormalizable kernel"
                )
            else:
                # the kernel's sum is near-zero, so it can't be scaled
                kernel_scale = 1
                normalized_kernel = kernel
        else:
            # the kernel is normalizable; we'll temporarily normalize it
            # now and undo the normalization later.
            normalized_kernel = kernel / kernel_scale

    if boundary is None:
        warnings.warn(
            "The convolve_fft version of boundary=None is "
            "equivalent to the convolve boundary='fill'.  There is "
            "no FFT equivalent to convolve's "
            "zero-if-kernel-leaves-boundary",
            AstropyUserWarning,
        )
        if psf_pad is None:
            psf_pad = True
        if fft_pad is None:
            fft_pad = True
    elif boundary == "fill":
        # create a boundary region at least as large as the kernel
        if psf_pad is False:
            warnings.warn(
                f"psf_pad was set to {psf_pad}, which overrides the "
                "boundary='fill' setting.",
                AstropyUserWarning,
            )
        else:
            psf_pad = True
        if fft_pad is None:
            # default is 'True' according to the docstring
            fft_pad = True
    elif boundary == "wrap":
        if psf_pad:
            raise ValueError("With boundary='wrap', psf_pad cannot be enabled.")
        psf_pad = False
        if fft_pad:
            raise ValueError("With boundary='wrap', fft_pad cannot be enabled.")
        fft_pad = False
        if dealias:
            raise ValueError("With boundary='wrap', dealias cannot be enabled.")
        fill_value = 0  # force zero; it should not be used
    elif boundary == "extend":
        raise NotImplementedError(
            "The 'extend' option is not implemented for fft-based convolution"
        )

    # Add shapes elementwise for psf_pad.
    if psf_pad:  # default=False
        # add the sizes along each dimension (bigger)
        newshape = np.array(arrayshape) + np.array(kernshape)
    else:
        # take the larger shape in each dimension (smaller)
        newshape = np.maximum(arrayshape, kernshape)

    if dealias:
        # Extend shape by 1/2 for dealiasing
        newshape += np.ceil(newshape / 2).astype(int)

    # Find ideal size for fft (was power of 2, now any powers of prime factors 2, 3, 5).
    if fft_pad:  # default=True
        # Get optimized sizes from scipy.
        newshape = _next_fast_lengths(newshape)

    # perform a second check after padding
    array_size_C = (
        np.prod(newshape, dtype=np.int64) * np.dtype(complex_dtype).itemsize
    ) * u.byte
    if array_size_C > 1 * u.GB and not allow_huge:
        raise ValueError(
            f"Size Error: Arrays will be {human_file_size(array_size_C)}.  "
            "Use allow_huge=True to override this exception."
        )

    # For future reference, this can be used to predict "almost exactly"
    # how much *additional* memory will be used.
    # size * (array + kernel + kernelfft + arrayfft +
    #         (kernel*array)fft +
    #         optional(weight image + weight_fft + weight_ifft) +
    #         optional(returned_fft))
    # total_memory_used_GB = (np.prod(newshape)*np.dtype(complex_dtype).itemsize
    #                        * (5 + 3*((interpolate_nan or ) and kernel_is_normalized))
    #                        + (1 + (not return_fft)) *
    #                          np.prod(arrayshape)*np.dtype(complex_dtype).itemsize
    #                        + np.prod(arrayshape)*np.dtype(bool).itemsize
    #                        + np.prod(kernshape)*np.dtype(bool).itemsize)
    #                        ) / 1024.**3

    # separate each dimension by the padding size...  this is to determine the
    # appropriate slice size to get back to the input dimensions
    arrayslices = []
    kernslices = []
    for newdimsize, arraydimsize, kerndimsize in zip(newshape, arrayshape, kernshape):
        center = newdimsize - (newdimsize + 1) // 2
        arrayslices += [
            slice(center - arraydimsize // 2, center + (arraydimsize + 1) // 2)
        ]
        kernslices += [
            slice(center - kerndimsize // 2, center + (kerndimsize + 1) // 2)
        ]
    arrayslices = tuple(arrayslices)
    kernslices = tuple(kernslices)

    if not np.all(newshape == arrayshape):
        if np.isfinite(fill_value):
            bigarray = np.ones(newshape, dtype=complex_dtype) * fill_value
        else:
            bigarray = np.zeros(newshape, dtype=complex_dtype)
        bigarray[arrayslices] = array
    else:
        bigarray = array

    if not np.all(newshape == kernshape):
        bigkernel = np.zeros(newshape, dtype=complex_dtype)
        bigkernel[kernslices] = normalized_kernel
    else:
        bigkernel = normalized_kernel

    arrayfft = fftn(bigarray)
    # need to shift the kernel so that, e.g., [0,0,1,0] -> [1,0,0,0] = unity
    kernfft = fftn(np.fft.ifftshift(bigkernel))
    fftmult = arrayfft * kernfft

    interpolate_nan = nan_treatment == "interpolate"
    if interpolate_nan:
        if not np.isfinite(fill_value):
            bigimwt = np.zeros(newshape, dtype=complex_dtype)
        else:
            bigimwt = np.ones(newshape, dtype=complex_dtype)

        bigimwt[arrayslices] = 1.0 - nanmaskarray * interpolate_nan
        wtfft = fftn(bigimwt)

        # You can only get to this point if kernel_is_normalized
        wtfftmult = wtfft * kernfft
        wtsm = ifftn(wtfftmult)
        # need to re-zero weights outside of the image (if it is padded, we
        # still don't weight those regions)
        bigimwt[arrayslices] = wtsm.real[arrayslices]
    else:
        bigimwt = 1

    if np.isnan(fftmult).any():
        # this check should be unnecessary; call it an insanity check
        raise ValueError("Encountered NaNs in convolve.  This is disallowed.")

    fftmult *= kernel_scale

    if array_unit is not None:
        fftmult <<= array_unit

    if return_fft:
        return fftmult

    if interpolate_nan:
        with np.errstate(divide="ignore", invalid="ignore"):
            # divide by zeros are expected here; if the weight is zero, we want
            # the output to be nan or inf
            rifft = (ifftn(fftmult)) / bigimwt
        if not np.isscalar(bigimwt):
            if min_wt > 0.0:
                rifft[bigimwt < min_wt] = np.nan
            else:
                # Set anything with no weight to zero (taking into account
                # slight offsets due to floating-point errors).
                rifft[bigimwt < 10 * np.finfo(bigimwt.dtype).eps] = 0.0
    else:
        rifft = ifftn(fftmult)

    if preserve_nan:
        rifft[arrayslices][nanmaskarray] = np.nan

    return rifft[arrayslices].real if crop else rifft.real


def interpolate_replace_nans(array, kernel, convolve=convolve, **kwargs):
    """
    Given a data set containing NaNs, replace the NaNs by interpolating from
    neighboring data points with a given kernel.

    Parameters
    ----------
    array : `numpy.ndarray`
        Array to be convolved with ``kernel``.  It can be of any
        dimensionality, though only 1, 2, and 3d arrays have been tested.
    kernel : `numpy.ndarray` or `astropy.convolution.Kernel`
        The convolution kernel. The number of dimensions should match those
        for the array.  The dimensions *do not* have to be odd in all directions,
        unlike in the non-fft `convolve` function.  The kernel will be
        normalized if ``normalize_kernel`` is set.  It is assumed to be centered
        (i.e., shifts may result if your kernel is asymmetric).  The kernel
        *must be normalizable* (i.e., its sum cannot be zero).
    convolve : `convolve` or `convolve_fft`
        One of the two convolution functions defined in this package.

    Returns
    -------
    newarray : `numpy.ndarray`
        A copy of the original array with NaN pixels replaced with their
        interpolated counterparts
    """
    if not np.any(np.isnan(array)):
        return array.copy()

    newarray = array.copy()

    convolved = convolve(
        array,
        kernel,
        nan_treatment="interpolate",
        normalize_kernel=True,
        preserve_nan=False,
        **kwargs,
    )

    isnan = np.isnan(array)
    newarray[isnan] = convolved[isnan]

    return newarray


def convolve_models(model, kernel, mode="convolve_fft", **kwargs):
    """
    Convolve two models using `~astropy.convolution.convolve_fft`.

    Parameters
    ----------
    model : `~astropy.modeling.core.Model`
        Functional model
    kernel : `~astropy.modeling.core.Model`
        Convolution kernel
    mode : str
        Keyword representing which function to use for convolution.

        * 'convolve_fft' : use `~astropy.convolution.convolve_fft` function.
        * 'convolve' : use `~astropy.convolution.convolve`.
    **kwargs : dict
        Keyword arguments to me passed either to `~astropy.convolution.convolve`
        or `~astropy.convolution.convolve_fft` depending on ``mode``.

    Returns
    -------
    default : `~astropy.modeling.core.CompoundModel`
        Convolved model
    """
    if mode == "convolve_fft":
        operator = SPECIAL_OPERATORS.add(
            "convolve_fft", partial(convolve_fft, **kwargs)
        )
    elif mode == "convolve":
        operator = SPECIAL_OPERATORS.add("convolve", partial(convolve, **kwargs))
    else:
        raise ValueError(f"Mode {mode} is not supported.")

    return CompoundModel(operator, model, kernel)


def convolve_models_fft(model, kernel, bounding_box, resolution, cache=True, **kwargs):
    """
    Convolve two models using `~astropy.convolution.convolve_fft`.

    Parameters
    ----------
    model : `~astropy.modeling.core.Model`
        Functional model
    kernel : `~astropy.modeling.core.Model`
        Convolution kernel
    bounding_box : tuple
        The bounding box which encompasses enough of the support of both
        the ``model`` and ``kernel`` so that an accurate convolution can be
        computed.
    resolution : float
        The resolution that one wishes to approximate the convolution
        integral at.
    cache : optional, bool
        Default value True. Allow for the storage of the convolution
        computation for later reuse.
    **kwargs : dict
        Keyword arguments to be passed either to `~astropy.convolution.convolve`
        or `~astropy.convolution.convolve_fft` depending on ``mode``.

    Returns
    -------
    default : `~astropy.modeling.core.CompoundModel`
        Convolved model
    """
    operator = SPECIAL_OPERATORS.add("convolve_fft", partial(convolve_fft, **kwargs))

    return Convolution(operator, model, kernel, bounding_box, resolution, cache)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.stats import (
    calculate_bin_edges,
    freedman_bin_width,
    histogram,
    knuth_bin_width,
    scott_bin_width,
)
from astropy.utils.compat.optional_deps import HAS_SCIPY


def test_scott_bin_width():
    rng = np.random.default_rng(0)
    N = 1000
    X = rng.standard_normal(N)

    delta = scott_bin_width(X)
    assert_allclose(delta, 3.5 * np.std(X) / N ** (1 / 3))

    delta, bins = scott_bin_width(X, return_bins=True)
    assert_allclose(delta, 3.5 * np.std(X) / N ** (1 / 3))

    with pytest.raises(ValueError):
        scott_bin_width(rng.random((2, 10)))


def test_freedman_bin_width():
    rng = np.random.default_rng(0)
    N = 10_000
    X = rng.standard_normal(N)

    v25, v75 = np.percentile(X, [25, 75])

    delta = freedman_bin_width(X)
    assert_allclose(delta, 2 * (v75 - v25) / N ** (1 / 3))

    delta, bins = freedman_bin_width(X, return_bins=True)
    assert_allclose(delta, 2 * (v75 - v25) / N ** (1 / 3))

    with pytest.raises(ValueError):
        freedman_bin_width(rng.random((2, 10)))

    # data with too small IQR
    test_x = [1, 2, 3] + [4] * 100 + [5, 6, 7]
    with pytest.raises(ValueError, match=r"Please use another bin method"):
        with pytest.warns(RuntimeWarning, match=r"divide by zero encountered"):
            freedman_bin_width(test_x, return_bins=True)

    # data with small IQR but not too small
    test_x = np.asarray([1, 2, 3] * 100 + [4] + [5, 6, 7], dtype=np.float32)
    test_x *= 1.5e-6
    delta, bins = freedman_bin_width(test_x, return_bins=True)
    assert_allclose(delta, 8.923325554510689e-07)


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_knuth_bin_width():
    rng = np.random.default_rng(0)
    X = rng.standard_normal(10_000)

    dx, bins = knuth_bin_width(X, return_bins=True)
    assert_allclose(len(bins), 58)

    dx2 = knuth_bin_width(X)
    assert dx == dx2

    with pytest.raises(ValueError):
        knuth_bin_width(rng.random((2, 10)))


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_knuth_histogram():
    rng = np.random.default_rng(0)
    x = rng.standard_normal(1000)
    counts, bins = histogram(x, "knuth")
    assert counts.sum() == len(x)
    assert len(counts) == len(bins) - 1


_bin_types_to_test = [30, "scott", "freedman", "blocks"]

if HAS_SCIPY:
    _bin_types_to_test += ["knuth"]


@pytest.mark.parametrize("bin_type", _bin_types_to_test + [np.linspace(-5, 5, 31)])
def test_histogram(bin_type):
    rng = np.random.default_rng(0)
    x = rng.standard_normal(1000)
    counts, bins = histogram(x, bin_type)
    assert counts.sum() == len(x)
    assert len(counts) == len(bins) - 1


# Don't include a list of bins as a bin_type here because the effect
# of range is different in that case
@pytest.mark.parametrize("bin_type", _bin_types_to_test)
def test_histogram_range(bin_type):
    # Regression test for #8010
    rng = np.random.default_rng(0)
    x = rng.standard_normal(1000)
    range = (0.1, 0.8)

    bins = calculate_bin_edges(x, bin_type, range=range)
    assert bins.max() == range[1]
    assert bins.min() == range[0]


def test_histogram_range_with_bins_list():
    # The expected result when the input bins is a list is
    # the same list on output.
    rng = np.random.default_rng(0)
    x = rng.standard_normal(1000)
    range = (0.1, 0.8)

    input_bins = np.linspace(-5, 5, 31)
    bins = calculate_bin_edges(x, input_bins, range=range)
    assert all(bins == input_bins)


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_histogram_output_knuth():
    rng = np.random.default_rng(0)
    X = rng.standard_normal(100)

    counts, bins = histogram(X, bins="knuth")
    assert_allclose(counts, [2, 1, 13, 19, 15, 18, 14, 10, 8])

    # fmt: off
    assert_allclose(bins, [-2.32503077, -1.84420596, -1.36338114, -0.88255632, -0.4017315,
                           0.07909331, 0.55991813, 1.04074295, 1.52156777, 2.00239258])
    # fmt: on


def test_histogram_output():
    rng = np.random.default_rng(0)
    X = rng.standard_normal(100)

    counts, bins = histogram(X, bins=10)
    assert_allclose(counts, [2, 0, 12, 14, 14, 17, 16, 8, 9, 8])

    # fmt: off
    assert_allclose(bins, [-2.32503077, -1.89228844, -1.4595461, -1.02680377, -0.59406143,
                           -0.1613191, 0.27142324, 0.70416558, 1.13690791, 1.56965025,
                           2.00239258])
    # fmt: on

    counts, bins = histogram(X, bins="scott")
    assert_allclose(counts, [2, 14, 27, 25, 16, 16])

    # fmt: off
    assert_allclose(bins, [-2.32503077, -1.59953424, -0.87403771, -0.14854117, 0.57695536,
                           1.3024519, 2.02794843])
    # fmt: on

    counts, bins = histogram(X, bins="freedman")
    assert_allclose(counts, [2, 11, 16, 18, 22, 14, 13, 4])

    # fmt: off
    assert_allclose(bins, [-2.32503077, -1.74087192, -1.15671306, -0.5725542,  0.01160465,
                           0.59576351, 1.17992237, 1.76408122, 2.34824008], rtol=2e-7)
    # fmt: on

    counts, bins = histogram(X, bins="blocks")
    assert_allclose(counts, [3, 97])
    assert_allclose(bins, [-2.32503077, -1.37136996, 2.00239258])


def test_histogram_badargs():
    x = np.array([1.0, 2.0, 3.0])

    # weights is not supported
    for bins in ["scott", "freedman", "blocks"]:
        with pytest.raises(NotImplementedError):
            histogram(x, bins, weights=x)

    # bad bins arg gives ValueError
    with pytest.raises(ValueError):
        histogram(x, bins="bad_argument")
